# Copyright 2018-2018 Berk Onat
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import numpy as np
from .CHARMMCommon import PARSERNAME, PROGRAMNAME, PARSERTAG
from nomadcore.smart_parser.SmartParserDictionary import metaNameConverter, metaNameConverter_UnderscoreSpaceDash
from nomadcore.smart_parser.SmartParserDictionary import MetaInfoMap, FileInfoMap, MapDictionary
from nomadcore.smart_parser.SmartParserDictionary import getDict_MetaStrInDict, getList_MetaStrInDict, get_unitDict
import nomadcore.md_data_access.MDDataAccess as MDData

def get_fileListDict():
    """Loads dictionary for file namelists.

    Returns:
        the list of defaults file namelists
    """

    startpage = {
        'nameTranslate'   : metaNameConverter,
        'metaHeader'      : PARSERTAG,
        'metaNameTag'     : 'inout_file',
        'metaInfoType'    : 'C',
        'activeMetaNames' : [],
        'activeSections'  : [PARSERTAG+'_section_input_output_files'],
        'fileInterface'   : ["charmmcoor", "parmed", "mdtraj", "mdanalysis"]
        #'fileInterface'   : ["charmmcoor", "parmed", "mdtraj", "mdanalysis", "pymolfile"]
        #'fileInterface'   : ["charmmcoor", "parmed", "pymolfile", "mdanalysis", "mdtraj"]
        }
    namelist = {
        'structure'     : FileInfoMap(startpage, fileFormat=['.psf'], activeInfo=True,
                                      infoPurpose=['topology', 'unitcell']),
        'trajectory'    : FileInfoMap(startpage, fileFormat=['.dcd'], activeInfo=True,
                                      infoPurpose=['trajectory', 'unitcell']),
        'traj_vel'      : FileInfoMap(startpage, fileFormat=['.dcd'], activeInfo=True,
                                      infoPurpose=['velocities']),
        'traj_force'    : FileInfoMap(startpage, fileFormat=['.xyz'], activeInfo=True,
                                      infoPurpose=['forces']),
        'output_coord'  : FileInfoMap(startpage, fileFormat=['.charmmcor'], activeInfo=False,
                                      infoPurpose=['outputcoordinates']),
        'out_coor_str'  : FileInfoMap(startpage, fileFormat=['.charmmstream'], activeInfo=False,
                                      infoPurpose=['outputcoordinates'], strDict=None),
        'output_vel'    : FileInfoMap(startpage, fileFormat=['.charmmcor'], activeInfo=False,
                                      infoPurpose=['outputvelocities']),
        'output_force'  : FileInfoMap(startpage, fileFormat=['.charmmcor'], activeInfo=False,
                                      infoPurpose=['outputforces']),
        'input_coord'   : FileInfoMap(startpage, fileFormat=['.charmmcor'],
                                      infoPurpose=['inputcoordinates']),
        'in_coor_str'   : FileInfoMap(startpage, fileFormat=['.charmmstream'],
                                      infoPurpose=['inputcoordinates'], strDict=None),
        'input_vel'     : FileInfoMap(startpage, fileFormat=['.charmmcor'],
                                      infoPurpose=['inputvelocities']),
        'restart_coord' : FileInfoMap(startpage, fileFormat=['.charmmcor'],
                                      infoPurpose=['inputcoordinates', 'outputcoordinates', 'unitcell']),
        'restart_vel'   : FileInfoMap(startpage, fileFormat=['.charmmcor'],
                                      infoPurpose=['inputvelocities', 'outputvelocities']),
        'rtf_file'      : FileInfoMap(startpage, fileFormat=['.charmmtop'],
                                      infoPurpose=['topology']),
        'prm_file'      : FileInfoMap(startpage, fileFormat=['.charmmpar'],
                                      infoPurpose=['topology']),
        'cor_file'      : FileInfoMap(startpage, fileFormat=['.charmmcor'],
                                      infoPurpose=['topology', 'inputcoordinates']),
        'stream'        : FileInfoMap(startpage, fileFormat=['.charmmstream'],
                                      infoPurpose=['topology'], strDict=None),
        'str_rtf'       : FileInfoMap(startpage, fileFormat=['.charmmstrrtf'],
                                      infoPurpose=['topology'], strDict=None),
        'str_par'       : FileInfoMap(startpage, fileFormat=['.charmmstrpar'],
                                      infoPurpose=['topology'], strDict=None),
        'output_log'    : FileInfoMap(startpage),
        'control_input' : FileInfoMap(startpage),
        }
    return MapDictionary(namelist)

def get_nameListDict(deflist):
    """Loads control in data of CHARMM.

    Args:
        deflist: name list definition (cntrl/parm).

    matchWith parameters:
        EOL = End of line
        PL  = Match with previous line
        PW  = Previous word
        NW  = Next word (until space, comma) (DEFAULT)
        UD  = until delimeter (can be any string/character)

        The text matched inside apostrophe or qoutation is extracted

    Returns:
        the list of namelists
    """
    startpage = {
        'nameTranslate'   : metaNameConverter_UnderscoreSpaceDash,
        'metaHeader'      : PARSERTAG,
        'metaNameTag'     : 'inout_control',
        'metaInfoType'    : 'C',
        'activeMetaNames' : [],
        'activeSections'  : [PARSERTAG+'_section_control_parameters'],
        }

    cntrllist = {
        'CRYSTAL' : MetaInfoMap(startpage, defaultValue=None),
        'Crystal Type =' : MetaInfoMap(startpage, defaultValue=None, replaceTag='Crystal Type'),
        'A     =' : MetaInfoMap(startpage, defaultValue=None, replaceTag='a length'),
        'B     =' : MetaInfoMap(startpage, defaultValue=None, replaceTag='b length'),
        'C     =' : MetaInfoMap(startpage, defaultValue=None, replaceTag='c length'),
        'Alpha =' : MetaInfoMap(startpage, defaultValue=None, replaceTag='Alpha'),
        'Beta ='  : MetaInfoMap(startpage, defaultValue=None, replaceTag='Beta'),
        'Gamma =' : MetaInfoMap(startpage, defaultValue=None, replaceTag='Gamma'),
        'MINI'    : MetaInfoMap(startpage, defaultValue=None),
        'NSTEP'   : MetaInfoMap(startpage, defaultValue=None),
        'INBFRQ'  : MetaInfoMap(startpage, defaultValue=None),
        ' STEP '  : MetaInfoMap(startpage, defaultValue=None, replaceTag='STEP'),
        'PRTMIN'  : MetaInfoMap(startpage, defaultValue=None),
        'TOLFUN'  : MetaInfoMap(startpage, defaultValue=None),
        'TOLGRD'  : MetaInfoMap(startpage, defaultValue=None),
        'TOLITR'  : MetaInfoMap(startpage, defaultValue=None),
        'TOLSTP'  : MetaInfoMap(startpage, defaultValue=None),
        'TFREQ'   : MetaInfoMap(startpage, defaultValue=None),
        ' PCUT '  : MetaInfoMap(startpage, defaultValue=None, replaceTag='PCUT'),
        'IHBFRQ'  : MetaInfoMap(startpage, defaultValue=None),
        'NCGCYC'  : MetaInfoMap(startpage, defaultValue=None, matchFirst=4),
        'NPRINT'  : MetaInfoMap(startpage, defaultValue=None, matchFirst=4),
        'NONBOND OPTION FLAGS:'    : MetaInfoMap(startpage, defaultValue=None,
            appendMatchesList=[('ELEC', 'MW'), ('VDW', 'MW'), ('ATOM', 'MW'),
                               ('RDIE', 'MW'), ('SWIT', 'MW'), ('VATO', 'MW'),
                               ('VSWI', 'MW'), ('BYGR', 'MW'), ('NOEX', 'MW'),
                               ('NOEW', 'MW')],
            appendMatchesUntil='CUTNB  =', appendMaxLines=2, addAsList=True),
        'CUTNB'   : MetaInfoMap(startpage, defaultValue=None),
        'CTEXNB'  : MetaInfoMap(startpage, defaultValue=None),
        'CTONNB'  : MetaInfoMap(startpage, defaultValue=None),
        'CTOFNB'  : MetaInfoMap(startpage, defaultValue=None),
        'CGONNB'  : MetaInfoMap(startpage, defaultValue=None),
        'CGOFNB'  : MetaInfoMap(startpage, defaultValue=None),
        'WMIN'    : MetaInfoMap(startpage, defaultValue=None),
        'CDIE'    : MetaInfoMap(startpage, defaultValue=None),
        'SWITCH'  : MetaInfoMap(startpage, defaultValue=None, matchFirst=4),
        'VSWITCH' : MetaInfoMap(startpage, defaultValue=None, matchFirst=4),
        'ATOMS'   : MetaInfoMap(startpage, defaultValue=None, matchFirst=4, matchAlso='CHARMM>'),
        'WRNMXD'  : MetaInfoMap(startpage, defaultValue=None),
        'E14FAC'  : MetaInfoMap(startpage, defaultValue=None),
        ' EPS '   : MetaInfoMap(startpage, defaultValue=None, replaceTag='EPS'),
        'NBXMOD'  : MetaInfoMap(startpage, defaultValue=None),
        'CUToff Hydrogen Bond  distance' : MetaInfoMap(startpage, defaultValue=None,
                                                       replaceTag='hydrogen bond cutoff distance'),
        '   Angle =' : MetaInfoMap(startpage, defaultValue=None,
                                   replaceTag='hydrogen bond cutoff angle'),
        'CuT switching ON HB dist.' : MetaInfoMap(startpage, defaultValue=None,
                                                  replaceTag='hydrogen bond switching on distance'),
        'OFf HB dist.' : MetaInfoMap(startpage, defaultValue=None,
                                     replaceTag='hydrogen bond switching off distance'),
        'CuT switching ON Hb Angle' : MetaInfoMap(startpage, defaultValue=None,
                                                  replaceTag='hydrogen bond switching on angle'),
        'OFf Hb Angle' : MetaInfoMap(startpage, defaultValue=None,
                                     replaceTag='hydrogen bond switching off angle'),
        'ACCEptor antecedents' : MetaInfoMap(startpage, defaultValue=None),
        'HBFIND-exclusions' : MetaInfoMap(startpage, defaultValue=None,
                                          replaceTag='hbond exclusions due to distance cutoff'),
        'due to distance cutoff,' : MetaInfoMap(startpage, defaultValue=None,
                                                replaceTag='hbond exclusions due to angle cutoff'),
        'HBEDIT-deletions' : MetaInfoMap(startpage, defaultValue=None,
                                         replaceTag='hbond deletions due to duplications'),
        'due to duplications,' : MetaInfoMap(startpage, defaultValue=None,
                                             replaceTag='hbond deletions due to best option'),
        'due to fixed atoms' : MetaInfoMap(startpage, defaultValue=None, matchWith='PW',
                                           replaceTag='hbond deletions due to fixed atoms'),
        'due to exclusions' : MetaInfoMap(startpage, defaultValue=None, matchWith='PW',
                                          replaceTag='hbond deletions due to exclusion'),
        'HBEDIT: currently' : MetaInfoMap(startpage, defaultValue=None,
                                          replaceTag='hydrogen bonds present'),
        'NCGCYC'  : MetaInfoMap(startpage, defaultValue=None),
        'Minimization exiting with' : MetaInfoMap(startpage, defaultValue=None,
                matchWith='EOL', replaceTag='minimization exit status'),
        'DYNA'    : MetaInfoMap(startpage, defaultValue=None),
        'AKMASTP' : MetaInfoMap(startpage, defaultValue=None),
        'FIRSTT'  : MetaInfoMap(startpage, defaultValue=None),
        'ISEED'   : MetaInfoMap(startpage, defaultValue=None),
        'IPRFRQ'  : MetaInfoMap(startpage, defaultValue=None),
        'IHTFRQ'  : MetaInfoMap(startpage, defaultValue=None),
        'IEQFRQ'  : MetaInfoMap(startpage, defaultValue=None),
        'IUNREA'  : MetaInfoMap(startpage, defaultValue=None),
        'IUNWRI'  : MetaInfoMap(startpage, defaultValue=None),
        'IUNOS'   : MetaInfoMap(startpage, defaultValue=None),
        'IUNCRD'  : MetaInfoMap(startpage, defaultValue=None),
        'IUNVEL'  : MetaInfoMap(startpage, defaultValue=None),
        'IUNXYZ'  : MetaInfoMap(startpage, defaultValue=None),
        'KUNIT'   : MetaInfoMap(startpage, defaultValue=None),
        'NSAVC'   : MetaInfoMap(startpage, defaultValue=None),
        'NSAVV'   : MetaInfoMap(startpage, defaultValue=None),
        'NSAVX'   : MetaInfoMap(startpage, defaultValue=None),
        'ISCALE'  : MetaInfoMap(startpage, defaultValue=None),
        'ISCVEL'  : MetaInfoMap(startpage, defaultValue=None),
        'IASORS'  : MetaInfoMap(startpage, defaultValue=None),
        'IASVEL'  : MetaInfoMap(startpage, defaultValue=None),
        'ICHECW'  : MetaInfoMap(startpage, defaultValue=None),
        'NTRFRQ'  : MetaInfoMap(startpage, defaultValue=None),
        'ILBFRQ'  : MetaInfoMap(startpage, defaultValue=None),
        'IMGFRQ'  : MetaInfoMap(startpage, defaultValue=None),
        'ISVFRQ'  : MetaInfoMap(startpage, defaultValue=None),
        'NCYCLE'  : MetaInfoMap(startpage, defaultValue=None, matchFirst=4),
        'NSNOS'   : MetaInfoMap(startpage, defaultValue=None),
        'TEMINC'  : MetaInfoMap(startpage, defaultValue=None),
        'TSTRUC'  : MetaInfoMap(startpage, defaultValue=None),
        'FINALT'  : MetaInfoMap(startpage, defaultValue=None),
        'TWINDH'  : MetaInfoMap(startpage, defaultValue=None),
        'TWINDL'  : MetaInfoMap(startpage, defaultValue=None),
        'TIME STEP' : MetaInfoMap(startpage, defaultValue=None),
        'RANDOM NUM. GEN. SEED\(S\)' : MetaInfoMap(startpage, defaultValue=None, matchWith='EOL'),
        'NUMBER OF DEGREES OF FREEDOM' : MetaInfoMap(startpage, defaultValue=None),
        'GAUSSIAN OPTION                  IS' : MetaInfoMap(startpage, defaultValue=None),
        'VELOCITIES ASSIGNED AT TEMPERATURE' : MetaInfoMap(startpage, defaultValue=None),
        'SEEDS>'  : MetaInfoMap(startpage, defaultValue=None, matchWith='EOL'),
        'SHAKE TOLERANCE' : MetaInfoMap(startpage, defaultValue=None),
        'DYNA'    : MetaInfoMap(startpage, defaultValue=None),
        'STRT'    : MetaInfoMap(startpage, defaultValue=None),
        'REST'    : MetaInfoMap(startpage, defaultValue=None),
        'QREF'    : MetaInfoMap(startpage, defaultValue=None),
        'TREF'    : MetaInfoMap(startpage, defaultValue=None),
        'HOOVER'  : MetaInfoMap(startpage, defaultValue=None, matchFirst=4),
        'REFT'    : MetaInfoMap(startpage, defaultValue=None, matchFirst=4),
        'TMASS'   : MetaInfoMap(startpage, defaultValue=None),
        'PCONS'   : MetaInfoMap(startpage, defaultValue=None),
        'PMASS'   : MetaInfoMap(startpage, defaultValue=None),
        }

    filecntrllist = {
        'OPEN'  : MetaInfoMap(startpage, defaultValue=None),
        'READ'  : MetaInfoMap(startpage, defaultValue=None),
        'WRITE' : MetaInfoMap(startpage, defaultValue=None),
        'CARD'  : MetaInfoMap(startpage, defaultValue=None),
        'COOR'  : MetaInfoMap(startpage, defaultValue=None),
        'FILE'  : MetaInfoMap(startpage, defaultValue=None),
        'UNIT'  : MetaInfoMap(startpage, defaultValue=None),
        'NAME'  : MetaInfoMap(startpage, defaultValue=None),
        'Parameter:' : MetaInfoMap(startpage, defaultValue=None, matchWith='EOL'),
        }

    readcntrllist = {
        'READ'  : MetaInfoMap(startpage, defaultValue=None),
        'WRITE' : MetaInfoMap(startpage, defaultValue=None),
        'CARD'  : MetaInfoMap(startpage, defaultValue=None),
        'COOR'  : MetaInfoMap(startpage, defaultValue=None),
        'FILE'  : MetaInfoMap(startpage, defaultValue=None),
        'UNIT'  : MetaInfoMap(startpage, defaultValue=None),
        }

    startpage.update({
        'metaNameTag'     : 'mdout',
        'activeSections'  : ['section_single_configuration_calculation']
        })
    mddatalist = {
        'CYCLE'     : MetaInfoMap(startpage, matchFirst=4),
        'EVAL'      : MetaInfoMap(startpage, matchLast=4),
        'STEP'      : MetaInfoMap(startpage, matchFirst=4),
        'ENERGY'    : MetaInfoMap(startpage, matchFirst=4),
        'TOTENER'   : MetaInfoMap(startpage, matchFirst=4),
        'TOTKE'     : MetaInfoMap(startpage, matchFirst=4),
        'TEMPERATURE' : MetaInfoMap(startpage, matchFirst=4),
        'DELTA-E'   : MetaInfoMap(startpage, matchFirst=4),
        'GRMS'      : MetaInfoMap(startpage, matchFirst=4),
        'HFCTOTE'   : MetaInfoMap(startpage, matchFirst=4),
        'HFCKE'     : MetaInfoMap(startpage, matchFirst=4),
        'EHFCOR'    : MetaInfoMap(startpage, matchFirst=4),
        'VIRKE'     : MetaInfoMap(startpage, matchFirst=4),
        'STEP-SIZE' : MetaInfoMap(startpage),
        'BONDS'     : MetaInfoMap(startpage, matchFirst=4),
        'ANGLES'    : MetaInfoMap(startpage, matchFirst=4),
        'UREY-B'    : MetaInfoMap(startpage, matchFirst=4),
        'DIHEDRALS' : MetaInfoMap(startpage, matchFirst=4),
        'IMPROPERS' : MetaInfoMap(startpage, matchFirst=4),
        'VDWAALS'   : MetaInfoMap(startpage, matchFirst=3),
        'ELEC'      : MetaInfoMap(startpage, matchFirst=4),
        'HBONDS'    : MetaInfoMap(startpage, matchFirst=4),
        'ASP'       : MetaInfoMap(startpage),
        'USER'      : MetaInfoMap(startpage),
        'Virial-Tensor' : MetaInfoMap(startpage),
        'Pressure-Tensor' : MetaInfoMap(startpage),
        'TIME'      : MetaInfoMap(startpage, matchFirst=4),
        'VIRI'      : MetaInfoMap(startpage, matchFirst=4),
        'VIRE'      : MetaInfoMap(startpage, matchFirst=4),
        'PRESSE'    : MetaInfoMap(startpage, matchFirst=4),
        'PRESSI'    : MetaInfoMap(startpage, matchFirst=4),
        'VOLUME'    : MetaInfoMap(startpage, matchFirst=4),
        'BPRE'      : MetaInfoMap(startpage, matchFirst=4),
        'VTOT'      : MetaInfoMap(startpage, matchFirst=4),
        'VKIN'      : MetaInfoMap(startpage, matchFirst=4),
        'EHYS'      : MetaInfoMap(startpage, matchFirst=4),
        'PRSE'      : MetaInfoMap(startpage, matchFirst=4),
        'PRSI'      : MetaInfoMap(startpage, matchFirst=4),
        'CMAP'      : MetaInfoMap(startpage, matchFirst=4),
        'STRB'      : MetaInfoMap(startpage, matchFirst=4),
        'OOPL'      : MetaInfoMap(startpage, matchFirst=4),
        'HARM'      : MetaInfoMap(startpage, matchFirst=4),
        'CDIH'      : MetaInfoMap(startpage, matchFirst=4),
        'CPUC'      : MetaInfoMap(startpage, matchFirst=4),
        'CIC'       : MetaInfoMap(startpage, matchFirst=4),
        'CDRO'      : MetaInfoMap(startpage, matchFirst=4),
        'NOE'       : MetaInfoMap(startpage, matchFirst=4),
        'SBOU'      : MetaInfoMap(startpage, matchFirst=4),
        'IMNB'      : MetaInfoMap(startpage, matchFirst=4),
        'EXTE'      : MetaInfoMap(startpage, matchFirst=4),
        'EWKS'      : MetaInfoMap(startpage, matchFirst=4),
        'EWSE'      : MetaInfoMap(startpage, matchFirst=4),
        'RXNF'      : MetaInfoMap(startpage, matchFirst=4),
        'ST2'       : MetaInfoMap(startpage, matchFirst=4),
        'IMST'      : MetaInfoMap(startpage, matchFirst=4),
        'TSM'       : MetaInfoMap(startpage, matchFirst=4),
        'QMEL'      : MetaInfoMap(startpage, matchFirst=4),
        'QMVD'      : MetaInfoMap(startpage, matchFirst=4),
        'EHAR'      : MetaInfoMap(startpage, matchFirst=4),
        'GEO'       : MetaInfoMap(startpage, matchFirst=4),
        'MDIP'      : MetaInfoMap(startpage, matchFirst=4),
        'PRMS'      : MetaInfoMap(startpage, matchFirst=4),
        'PANG'      : MetaInfoMap(startpage, matchFirst=4),
        'SSBP'      : MetaInfoMap(startpage, matchFirst=4),
        'BK4D'      : MetaInfoMap(startpage, matchFirst=4),
        'SHEL'      : MetaInfoMap(startpage, matchFirst=4),
        'RESD'      : MetaInfoMap(startpage, matchFirst=4),
        'SHAP'      : MetaInfoMap(startpage, matchFirst=4),
        'PULL'      : MetaInfoMap(startpage, matchFirst=4),
        'POLA'      : MetaInfoMap(startpage, matchFirst=4),
        'DMC'       : MetaInfoMap(startpage, matchFirst=4),
        'RGY'       : MetaInfoMap(startpage, matchFirst=4),
        'EWEX'      : MetaInfoMap(startpage, matchFirst=4),
        'EWQC'      : MetaInfoMap(startpage, matchFirst=4),
        'EWUT'      : MetaInfoMap(startpage, matchFirst=4),
        'VEXX'      : MetaInfoMap(startpage, matchFirst=4),
        'VEXY'      : MetaInfoMap(startpage, matchFirst=4),
        'VEXZ'      : MetaInfoMap(startpage, matchFirst=4),
        'VEYX'      : MetaInfoMap(startpage, matchFirst=4),
        'VEYY'      : MetaInfoMap(startpage, matchFirst=4),
        'VEYZ'      : MetaInfoMap(startpage, matchFirst=4),
        'VEZX'      : MetaInfoMap(startpage, matchFirst=4),
        'VEZY'      : MetaInfoMap(startpage, matchFirst=4),
        'VEZZ'      : MetaInfoMap(startpage, matchFirst=4),
        'VIXX'      : MetaInfoMap(startpage, matchFirst=4),
        'VIXY'      : MetaInfoMap(startpage, matchFirst=4),
        'VIXZ'      : MetaInfoMap(startpage, matchFirst=4),
        'VIYX'      : MetaInfoMap(startpage, matchFirst=4),
        'VIYY'      : MetaInfoMap(startpage, matchFirst=4),
        'VIYZ'      : MetaInfoMap(startpage, matchFirst=4),
        'VIZX'      : MetaInfoMap(startpage, matchFirst=4),
        'VIZY'      : MetaInfoMap(startpage, matchFirst=4),
        'VIZZ'      : MetaInfoMap(startpage, matchFirst=4),
        'PEXX'      : MetaInfoMap(startpage, matchFirst=4),
        'PEXY'      : MetaInfoMap(startpage, matchFirst=4),
        'PEXZ'      : MetaInfoMap(startpage, matchFirst=4),
        'PEYX'      : MetaInfoMap(startpage, matchFirst=4),
        'PEYY'      : MetaInfoMap(startpage, matchFirst=4),
        'PEYZ'      : MetaInfoMap(startpage, matchFirst=4),
        'PEZX'      : MetaInfoMap(startpage, matchFirst=4),
        'PEZY'      : MetaInfoMap(startpage, matchFirst=4),
        'PEZZ'      : MetaInfoMap(startpage, matchFirst=4),
        'PIXX'      : MetaInfoMap(startpage, matchFirst=4),
        'PIXY'      : MetaInfoMap(startpage, matchFirst=4),
        'PIXZ'      : MetaInfoMap(startpage, matchFirst=4),
        'PIYX'      : MetaInfoMap(startpage, matchFirst=4),
        'PIYY'      : MetaInfoMap(startpage, matchFirst=4),
        'PIYZ'      : MetaInfoMap(startpage, matchFirst=4),
        'PIZX'      : MetaInfoMap(startpage, matchFirst=4),
        'PIZY'      : MetaInfoMap(startpage, matchFirst=4),
        'PIZZ'      : MetaInfoMap(startpage, matchFirst=4),
        }

    startpage.update({
        'metaNameTag'     : 'parm',
        'activeSections'  : [PARSERTAG+'_mdin_method']
        })
    extralist = {
        'flags' : MetaInfoMap(startpage),
        'number_of_atoms' : MetaInfoMap(startpage),
        'minimization' : MetaInfoMap(startpage, activeInfo=True),
        'integrator' : MetaInfoMap(startpage, activeInfo=True),
        'thermostat' : MetaInfoMap(startpage, activeInfo=True),
        'barostat' : MetaInfoMap(startpage, activeInfo=True),
        'box_info' : MetaInfoMap(startpage),
        'unitcell_radius' : MetaInfoMap(startpage),
        'total_memory' : MetaInfoMap(startpage),
        'file_format' : MetaInfoMap(startpage),
        'file_version' : MetaInfoMap(startpage),
        'file_date' : MetaInfoMap(startpage),
        'file_time' : MetaInfoMap(startpage)
        }

    if deflist == 'mddata':
        namelist = mddatalist
#    elif deflist == 'parm':
#        namelist = parmlist
    elif deflist == 'extra':
        namelist = extralist
    elif deflist == 'filecntrl':
        namelist = filecntrllist
    elif deflist == 'stepcontrol':
        namelist = stepcontrol
    else:
        namelist = cntrllist
    return MapDictionary(namelist)

def set_Dictionaries(self):
        self.unitDict = get_unitDict('si')
        self.akmaDict = get_unitDict('akma')
        self.fileDict = get_fileListDict()
        self.cntrlDict = get_nameListDict('cntrl')
        self.filecntrlDict = get_nameListDict('filecntrl')
        #self.parmDict = get_nameListDict('parm')
        self.mddataDict = get_nameListDict('mddata')
        self.extraDict = get_nameListDict('extra')
        self.mdoutHeaderDict = {}
        #self.fileUnitDict = {}
        self.stepcontrolDict = {
            'logsteps'       : None,
            'nextlogsteps'   : None,
            'trajsteps'      : None,
            'velsteps'       : None,
            'forcesteps'     : None,
            'steps'          : None,
            'follow'         : None,
            'sectioncontrol' : None,
            'cntrlparmstep'  : [-1],
        }
        self.metaDicts = {
            'file'   : self.fileDict,
            'cntrl'  : self.cntrlDict,
            'filecntrl'  : self.filecntrlDict,
#            'parm'   : self.parmDict,
            'mddata' : self.mddataDict,
            'extra'  : self.extraDict,
            }
        self.sectionDict = {
                'section' : {
                    "metaNameTag"      : ['input_output_files',
                                          'control_parameters'],
                    "listTypStr"       : 'type_section',
                    "repeatingSection" : False,
                    "supraNames"       : ["section_run"]
                    }
                }

def get_updateDictionary(self, defname):

    def value_convert_array(itemdict):
        val = itemdict["value"]
        if(val is not None and isinstance(val, str)):
            if("[" in val or "(" in val):
                return False, np.asarray(val), itemdict
        return False, val, itemdict

    def value_convert_array_norm(itemdict):
        val = []
        depdict = itemdict["depends"]
        for item in depdict["value"]:
            st=self.metaStorage.fetchAttrValue(item)
            if(st is not None and isinstance(st, str)):
                if("[" in st or "(" in st):
                    val.append(list(storedtext))
        if val:
            return False, np.linalg.norm(np.asarray(val)), itemdict
        else:
            return False, itemdict["value"], itemdict

    startpage = {
        'metaHeader'      : '',
        'metaNameTag'     : '',
        'activeMetaNames' : [],
        'activeSections'  : ['section_sampling_method','section_single_configuration_calculation',PARSERTAG+'_mdout']
        }

    # ---------------------------------------------------------------
    #   Definitions of meta data values for section_sampling_method
    # ---------------------------------------------------------------
    sampling = {
        #'ensemble_type' : MetaInfoMap(startpage, activeInfo=True,
        #    depends=[
        #        {'test' : [['DYNA', ' is not None'],
        #                   ['FIRSTT', ' in [\"tcouple\",'
        #                                  '\"rescale\",'
        #                                  '\"reassign\",'
        #                                  '\"andersen\",'
        #                                  '\"langevin\"]'],
        #                   ['barostat', ' in [\"berendsen\",'
        #                                 '\"langevin\"]']],
        #         'assign' : 'NPT'},
        #        {'test' : [['minimization', ' is None'],
        #                   ['thermostat', ' in [\"langevin\"]'],
        #                   ['thermostat', ' not in [\"tcouple\",'
        #                                  '\"rescale\",'
        #                                  '\"reassign\",'
        #                                  '\"andersen\"]'],
        #                   ['barostat', ' is None']],
        #         'assign' : 'Langevin'},
        #        {'test' : [['minimization', ' is None'],
        #                   ['thermostat', ' in [\"tcouple\",'
        #                                  '\"rescale\",'
        #                                  '\"reassign\",'
        #                                  '\"andersen\"]'],
        #                   ['barostat', ' is None']],
        #         'assign' : 'NVT'},
        #        {'test' : [['minimization', ' is None'],
        #                   ['thermostat', ' is None'],
        #                   ['barostat', ' is None']],
        #         'assign' : 'NVE'},
        #        {'test' : [['minimization', ' is not None']],
        #         'assign' : 'minimization'},
        #        ],
        #    lookupdict=self.extraDict
        #    ),
        'sampling_method' : MetaInfoMap(startpage, activeInfo=True,
            depends=[
                {'test' : [['MINI', ' is not None']],
                 'assign' : 'geometry_optimization'},
                {'test' : [['DYNA', ' is not None']],
                 'assign' : 'molecular_dynamics'}
                ],
            lookupdict=self.cntrlDict
            ),
#            'settings_geometry_optimization' : MetaInfoMap(startpage),
#            'settings_metadynamics' : MetaInfoMap(startpage),
#            'settings_molecular_dynamics' : MetaInfoMap(startpage),
#            'settings_monte_carlo' : MetaInfoMap(startpage),
#        'geometry_optimization_energy_change' : MetaInfoMap(startpage,
#            depends={
#                '' : {'imin' : '1'},
#                },
#            lookupdict=self.cntrlDict
#            ),
#       'geometry_optimization_geometry_change' : MetaInfoMap(startpage),
        'geometry_optimization_method' : MetaInfoMap(startpage,
            depends=[
                {'test' : [['MINI',' is not None'],
                           ['MINI', ' in [\"SD\",\"sd\"]']],
                 'assign' : 'SD'},
                {'test' : [['MINI',' is not None'],
                           ['MINI', ' in [\"ABNR\",\"abnr\"]']],
                 'assign' : 'Adopted Basis Newton-Raphson'},
                {'test' : [['MINI',' is not None'],
                           ['MINI', ' in [\"NRAP\",\"nrap\"]']],
                 'assign' : 'Newton-Raphson'},
                {'test' : [['MINI',' is not None'],
                           ['MINI', ' in [\"POWELL\",\"powell\"]']],
                 'assign' : 'Powell'},
                {'test' : [['MINI',' is not None'],
                           ['MINI', ' in [\"POWE\",\"powe\"]']],
                 'assign' : 'Powell'},
                {'test' : [['MINI',' is not None'],
                           ['MINI', ' in [\"TNPACK\",\"tnpack\"]']],
                 'assign' : 'Truncated Newton Method'},
                {'test' : [['MINI',' is not None'],
                           ['MINI', ' not in [\"SD\",\"sd\",'
                                             '\"ABNR\",\"abnr\",'
                                             '\"NRAP\",\"nrap\",'
                                             '\"POWE\",\"powe\",'
                                             '\"POWELL\",\"powell\",'
                                             '\"TNPACK\",\"tnpack\"]']],
                 'assign' : 'CG'},
                ],
            lookupdict=self.cntrlDict,
            activeInfo=True,
            ),
#       'geometry_optimization_threshold_force' : MetaInfoMap(startpage),
        #'x_charmm_barostat_type' : MetaInfoMap(startpage,
        #    depends=[
        #        {'test' : [['minimization', ' is None'],
        #                   ['barostat', '== \"berendsen\"']],
        #         'assign' : 'Berendsen'},
        #        {'test' : [['minimization', ' is None'],
        #                   ['barostat', '== \"langevin\"']],
        #         'assign' : 'Nose-Hoover Langevin'}
        #        ],
        #    lookupdict=self.extraDict,
        #    #autoSections=True,
        #    activeInfo=True,
        #    activeSections=['settings_barostat']
        #    ),
        #'x_charmm_barostat_target_pressure' : MetaInfoMap(startpage,
        #    depends=[
        #        {'test' : [['TARGET PRESSURE', ' is not None']],
        #         'value' : 'TARGET PRESSURE'}
        #        ],
        #    lookupdict=self.cntrlDict,
        #    valtype='float',
        #    unitdict=self.unitDict,
        #    unit='atmosphere',
        #    #autoSections=True,
        #    activeInfo=True,
        #    activeSections=['settings_barostat']
        #    ),
        #'x_charmm_barostat_tau' : MetaInfoMap(startpage,
        #    depends=[
        #        {'test' : [['LANGEVIN OSCILLATION PERIOD', ' is not None']],
        #         'value' : 'LANGEVIN OSCILLATION PERIOD'},
        #        {'test' : [['BERENDSEN RELAXATION TIME', ' is not None']],
        #         'value' : 'BERENDSEN RELAXATION TIME'}
        #        ],
        #    lookupdict=self.cntrlDict,
        #    valtype='float',
        #    unitdict=self.unitDict,
        #    unit='akmatime',
        #    #autoSections=True,
        #    activeInfo=True,
        #    activeSections=['settings_barostat']
        #    ),
        #'x_charmm_integrator_type' : MetaInfoMap(startpage,
        #    depends=[
        #        {'test' : [['integrator', ' is not None']],
        #         'value' : 'integrator'},
        #        {'test' : [['thermostat', ' == \"langevin\"']],
        #         'assign' : 'Langevin'}
        #        ],
        #    lookupdict=self.extraDict,
        #    #autoSections=True,
        #    activeInfo=True,
        #    activeSections=['settings_integrator']
        #    ),
        'x_charmm_integrator_dt' : MetaInfoMap(startpage,
            depends=[
                {'test' : [['DYNA', ' is not None'],
                           ['TIME STEP', ' is not None']],
                 'value' : 'TIME STEP'},
                ],
            lookupdict=self.cntrlDict,
            valtype='float',
            unitdict=self.unitDict,
            unit='akmatime',
            #autoSections=True,
            activeInfo=True,
            activeSections=['settings_integrator']
            ),
        'x_charmm_number_of_steps_requested' : MetaInfoMap(startpage,
            depends=[{'value' : 'NSTEP'}],
            lookupdict=self.cntrlDict,
            valtype='int',
            #autoSections=True,
            activeInfo=True,
            activeSections=['settings_integrator']
            ),
        #'x_charmm_thermostat_type' : MetaInfoMap(startpage,
        #    depends=[
        #        {'test' : [['minimization', ' is None'],
        #                   ['thermostat', ' == \"tcouple\"']],
        #         'assign' : 'Berendsen'},
        #        {'test' : [['minimization', ' is None'],
        #                   ['thermostat', ' == \"rescale\"']],
        #         'assign' : 'Velocity Rescaling'},
        #        {'test' : [['minimization', ' is None'],
        #                   ['thermostat', ' == \"reassign\"']],
        #         'assign' : 'Velocity Reassigning'},
        #        {'test' : [['minimization', ' is None'],
        #                   ['thermostat', ' == \"andersen\"']],
        #         'assign' : 'Andersen'},
        #        {'test' : [['minimization', ' is None'],
        #                   ['thermostat', ' == \"langevin\"']],
        #         'assign' : 'Nose-Hoover Langevin'},
        #        ],
        #    lookupdict=self.extraDict,
        #    #autoSections=True,
        #    activeInfo=True,
        #    activeSections=['settings_thermostat']
        #    ),
        'x_charmm_thermostat_target_temperature' : MetaInfoMap(startpage,
            depends=[
                {'test' : [['DYNA', ' is not None'],
                           ['FINALT', ' is not None']],
                 'value' : 'FINALT'},
                ],
            lookupdict=self.cntrlDict,
            valtype='float',
            unitdict=self.unitDict,
            unit='Kelvin',
            #autoSections=True,
            activeInfo=True,
            activeSections=['settings_thermostat']
            ),
        #'x_charmm_thermostat_tau' : MetaInfoMap(startpage,
        #    depends=[
        #        {'test' : [['MINIMIZATION', ' is None'],
        #                   ['VELOCITY QUENCHING', ' is None'],
        #                   ['VELOCITY RESCALE FREQ', ' is not None']],
        #         'value' : 'VELOCITY RESCALE FREQ'},
        #        {'test' : [['MINIMIZATION', ' is None'],
        #                   ['VELOCITY QUENCHING', ' is None'],
        #                   ['VELOCITY REASSIGNMENT FREQ', ' is not None']],
        #         'value' : 'VELOCITY REASSIGNMENT FREQ'},
        #        {'test' : [['MINIMIZATION', ' is None'],
        #                   ['VELOCITY QUENCHING', ' is None'],
        #                   ['LOWE-ANDERSEN RATE', ' is not None']],
        #         'value' : 'LOWE-ANDERSEN RATE'},
        #        ],
        #    lookupdict=self.cntrlDict,
        #    valtype='float',
        #    unitdict=self.unitDict,
        #    unit='pico-second',
        #    #autoSections=True,
        #    activeInfo=True,
        #    activeSections=['settings_thermostat']
        #    ),
        'x_charmm_periodicity_type' : MetaInfoMap(startpage,
            depends=[
                {'test' : [['CRYSTAL', ' is not None']],
                 'assign' : 'periodic boundaries'},
                {'test' : [['CRYSTAL',' is None']],
                 'assign' : 'no periodic boundaries'},
                ],
            lookupdict=self.cntrlDict,
            #autoSections=True,
            activeInfo=True,
            activeSections=['settings_thermostat']
            )
        }

    # ------------------------------------------------------------
    #   Definitions for section_single_configuration_calculation
    # ------------------------------------------------------------
    singleconfcalc = {
        #'atom_forces_type' : MetaInfoMap(startpage,
        #    depends=[{'assign' : 'Force Field'}],
        #    lookupdict=self.mddataDict
        #    ),
        'energy_correction_entropy' : MetaInfoMap(startpage),
        'energy_current' : MetaInfoMap(startpage),
        'energy_electrostatic' : MetaInfoMap(startpage,
            depends=[{'value' : 'ELEC'}],
            valtype='float',
            unitdict=self.unitDict,
            unit=self.akmaDict['energy'],
            activeInfo=True,
            lookupdict=self.mddataDict
            ),
        'energy_free_per_atom' : MetaInfoMap(startpage),
        'energy_free' : MetaInfoMap(startpage),
        #'energy_method_current' : MetaInfoMap(startpage,
        #    depends=[{'assign' : 'Force Field'}],
        #    lookupdict=self.mddataDict
        #    ),
        'energy_t0_per_atom' : MetaInfoMap(startpage),
        'energy_total_t0_per_atom' : MetaInfoMap(startpage),
        'energy_total_t0' : MetaInfoMap(startpage,
            depends=[{'value' : 'ENERGY'}],
            valtype='float',
            unitdict=self.unitDict,
            unit=self.akmaDict['energy'],
            activeInfo=True,
            lookupdict=self.mddataDict
            ),
        'energy_total' : MetaInfoMap(startpage,
            depends=[{'value' : 'TOTENER'}],
            valtype='float',
            unitdict=self.unitDict,
            unit=self.akmaDict['energy'],
            activeInfo=True,
            lookupdict=self.mddataDict
            ),
        'hessian_matrix' : MetaInfoMap(startpage),
        'single_configuration_calculation_converged' : MetaInfoMap(startpage),
        'single_configuration_calculation_to_system_ref' : MetaInfoMap(startpage),
        'single_configuration_calculation_to_method_ref' : MetaInfoMap(startpage),
        'time_calculation' : MetaInfoMap(startpage),
        'time_single_configuration_calculation_cpu1_end' : MetaInfoMap(startpage),
        'time_single_configuration_calculation_cpu1_start' : MetaInfoMap(startpage),
        'time_single_configuration_calculation_date_end' : MetaInfoMap(startpage),
        'time_single_configuration_calculation_date_start' : MetaInfoMap(startpage),
        'time_single_configuration_calculation_wall_end' : MetaInfoMap(startpage),
        'time_single_configuration_calculation_wall_start' : MetaInfoMap(startpage),
        'stress_tensor_kind' : MetaInfoMap(startpage,
            depends=[
                {'test' : [['MINI', ' is not None']],
                 'assign' : 'geometry_optimization'},
                {'test' : [['DYNA', ' is not None']],
                 'assign' : 'molecular_dynamics'}
                ],
            activeInfo=True,
            lookupdict=self.extraDict
            ),
        'stress_tensor_value' : MetaInfoMap(startpage)
        }

    # section_single_energy_van_der_Waals of section_single_configuration_calculation
    singlevdw = {
        'energy_van_der_waals_value' : MetaInfoMap(startpage,
            depends=[{'value' : 'VDWAALS'}],
            lookupdict=self.mddataDict,
            valtype='float',
            unitdict=self.unitDict,
            unit=self.akmaDict['energy'],
            #autoSections=True,
            activeSections=['section_energy_van_der_waals']
            ),
        }

    # Info for section_restricted_uri
    restrictions = {
        'number_of_restricted_uri_files' : MetaInfoMap(startpage,
            value=int(1),
            #autoSections=True,
            activeInfo=True,
            activeSections=['section_restricted_uri']
            ),
        'restricted_uri_files' : MetaInfoMap(startpage,
            #autoSections=True,
            activeInfo=True,
            activeSections=['section_restricted_uri'],
            subfunction={
                'function' : self.parameter_file_name,
                'supportDict' : self.fileDict,
                },
            ),
        'restricted_uri_license' : MetaInfoMap(startpage,
            value='CHARMM License',
            #autoSections=True,
            activeInfo=True,
            activeSections=['section_restricted_uri']
            ),
        'restricted_uri_restriction' : MetaInfoMap(startpage,
            value='any access',
            #autoSections=True,
            activeInfo=True,
            activeSections=['section_restricted_uri']
            ),
        'restricted_uri_reason' : MetaInfoMap(startpage,
            value='propriety license',
            #autoSections=True,
            activeInfo=True,
            activeSections=['section_restricted_uri']
            ),
        'restricted_uri_issue_authority' : MetaInfoMap(startpage,
            value='CHARMM',
            #autoSections=True,
            activeInfo=True,
            activeSections=['section_restricted_uri']
            ),
        }

    # ------------------------------------------
    #   Definitions for section_frame_sequence
    # ------------------------------------------
    frameseq = {
        #'xxx_to_rm_frame_sequence_conserved_quantity_frames' : MetaInfoMap(startpage,
        #    depends=[{'store' : 'TS'}],
        #    valtype='int',
        #    lookupdict=self.mddataDict
        #    ),
        #'frame_sequence_conserved_quantity_stats' : MetaInfoMap(startpage),
        #'xxx_conserved_quantity' : MetaInfoMap(startpage,
        #    depends=[{'store' : ''}],
        #    valtype='float',
        #    unitdict=self.unitDict,
        #    unit='kilo-joule/mol',
        #    lookupdict=self.mddataDict
        #    ),
        'frame_sequence_continuation_kind' : MetaInfoMap(startpage),
        'frame_sequence_external_url' : MetaInfoMap(startpage),
        'xxx_to_rm_frame_sequence_kinetic_energy_frames' : MetaInfoMap(startpage,
            depends=[{'store' : 'STEP'}],
            valtype='int',
            lookupdict=self.mddataDict
            ),
        'frame_sequence_kinetic_energy_stats' : MetaInfoMap(startpage),
        'xxx_kinetic_energy' : MetaInfoMap(startpage,
            depends=[{'store' : 'TOTKE'}],
            valtype='float',
            unitdict=self.unitDict,
            unit=self.akmaDict['energy'],
            lookupdict=self.mddataDict
            ),
        'frame_sequence_to_frames_ref' : MetaInfoMap(startpage),
        'xxx_to_rm_frame_sequence_potential_energy_frames' : MetaInfoMap(startpage,
            depends=[{'store' : 'STEP'}],
            valtype='int',
            lookupdict=self.mddataDict
            ),
        'frame_sequence_potential_energy_stats' : MetaInfoMap(startpage),
        'xxx_potential_energy' : MetaInfoMap(startpage,
            depends=[{'store' : 'ENERGY'}],
            valtype='float',
            unitdict=self.unitDict,
            unit=self.akmaDict['energy'],
            lookupdict=self.mddataDict
            ),
        #'frame_sequence_pressure_average_frames' : MetaInfoMap(startpage,
        #    depends=[{'store' : 'STEP'}],
        #    valtype='int',
        #    lookupdict=self.mddataDict
        #    ),
        #'frame_sequence_pressure_average' : MetaInfoMap(startpage,
        #    depends=[{'store' : 'PRSI'}],
        #    valtype='float',
        #    unitdict=self.unitDict,
        #    unit='atmosphere',
        #    lookupdict=self.mddataDict
        #    ),
        'xxx_to_rm_frame_sequence_pressure_frames' : MetaInfoMap(startpage,
            depends=[{'store' : 'STEP'}],
            valtype='int',
            lookupdict=self.mddataDict
            ),
        'frame_sequence_pressure_stats' : MetaInfoMap(startpage),
        'xxx_instant_pressure' : MetaInfoMap(startpage,
            depends=[{'store' : 'PRESSI'}],
            valtype='float',
            unitdict=self.unitDict,
            unit='atmosphere',
            lookupdict=self.mddataDict
            ),
        'x_charmm_frame_sequence_external_pressure_frames' : MetaInfoMap(startpage,
            depends=[{'store' : 'STEP'}],
            valtype='int',
            lookupdict=self.mddataDict
            ),
        'x_charmm_frame_sequence_external_pressure_stats' : MetaInfoMap(startpage),
        'x_charmm_frame_sequence_external_pressure' : MetaInfoMap(startpage,
            depends=[{'store' : 'PRESSE'}],
            valtype='float',
            unitdict=self.unitDict,
            unit='atmosphere',
            lookupdict=self.mddataDict
            ),
        'x_charmm_frame_sequence_internal_pressure_frames' : MetaInfoMap(startpage,
            depends=[{'store' : 'STEP'}],
            valtype='int',
            lookupdict=self.mddataDict
            ),
        'x_charmm_frame_sequence_internal_pressure_stats' : MetaInfoMap(startpage),
        'x_charmm_frame_sequence_internal_pressure' : MetaInfoMap(startpage,
            depends=[{'store' : 'PRESSI'}],
            valtype='float',
            unitdict=self.unitDict,
            unit='atmosphere',
            lookupdict=self.mddataDict
            ),
        #'frame_sequence_temperature_average_frames' : MetaInfoMap(startpage,
        #    depends=[{'store' : 'TS'}],
        #    valtype='int',
        #    lookupdict=self.mddataDict
        #    ),
        #'frame_sequence_temperature_average' : MetaInfoMap(startpage,
        #    depends=[{'store' : 'TEMPAVG'}],
        #    valtype='float',
        #    unitdict=self.unitDict,
        #    unit='Kelvin',
        #    lookupdict=self.mddataDict
        #    ),
        'xxx_to_rm_frame_sequence_temperature_frames' : MetaInfoMap(startpage,
            depends=[{'store' : 'STEP'}],
            valtype='int',
            lookupdict=self.mddataDict
            ),
        'frame_sequence_temperature_stats' : MetaInfoMap(startpage),
        'xxx_instant_temperature' : MetaInfoMap(startpage,
            depends=[{'store' : 'TEMPERATURE'}],
            valtype='float',
            unitdict=self.unitDict,
            unit='Kelvin',
            lookupdict=self.mddataDict
            ),
        'frame_sequence_time' : MetaInfoMap(startpage,
            depends=[{'store' : 'TIME'}],
            valtype='float',
            unitdict=self.unitDict,
            unit='pico-second',
            # Calculates in AKMA units including inputs but prints in picoseconds.
            lookupdict=self.mddataDict
            ),
        'x_charmm_frame_sequence_volume_frames' : MetaInfoMap(startpage,
            activeSections=['section_frame_sequence'],
            depends=[{'store' : 'STEP'}],
            valtype='int',
            lookupdict=self.mddataDict
            ),
        'x_charmm_frame_sequence_volume' : MetaInfoMap(startpage,
            activeSections=['section_frame_sequence'],
            depends=[{'store' : 'VOLUME'}],
            valtype='float',
            unitdict=self.unitDict,
            unit='Angstrom**3',
            lookupdict=self.mddataDict
            ),
        'x_charmm_frame_sequence_total_energy_frames' : MetaInfoMap(startpage,
            activeSections=['section_frame_sequence'],
            depends=[{'store' : 'STEP'}],
            valtype='int',
            lookupdict=self.mddataDict
            ),
        'x_charmm_frame_sequence_total_energy' : MetaInfoMap(startpage,
            activeSections=['section_frame_sequence'],
            depends=[{'store' : 'TOTENER'}],
            valtype='float',
            unitdict=self.unitDict,
            unit=self.akmaDict['energy'],
            lookupdict=self.mddataDict
            ),
        'x_charmm_frame_sequence_total_kinetic_energy_frames' : MetaInfoMap(startpage,
            activeSections=['section_frame_sequence'],
            depends=[{'store' : 'STEP'}],
            valtype='int',
            lookupdict=self.mddataDict
            ),
        'x_charmm_frame_sequence_total_kinetic_energy' : MetaInfoMap(startpage,
            activeSections=['section_frame_sequence'],
            depends=[{'store' : 'TOTKE'}],
            valtype='float',
            unitdict=self.unitDict,
            unit=self.akmaDict['energy'],
            lookupdict=self.mddataDict
            ),
        'x_charmm_frame_sequence_high_frequency_correction_total_energy_frames' : MetaInfoMap(startpage,
            activeSections=['section_frame_sequence'],
            depends=[{'store' : 'STEP'}],
            valtype='int',
            lookupdict=self.mddataDict
            ),
        'x_charmm_frame_sequence_high_frequency_correction_total_energy' : MetaInfoMap(startpage,
            activeSections=['section_frame_sequence'],
            depends=[{'store' : 'HFCTOTE'}],
            valtype='float',
            unitdict=self.unitDict,
            unit=self.akmaDict['energy'],
            lookupdict=self.mddataDict
            ),
        'x_charmm_frame_sequence_high_frequency_correction_kinetic_energy_frames' : MetaInfoMap(startpage,
            activeSections=['section_frame_sequence'],
            depends=[{'store' : 'STEP'}],
            valtype='int',
            lookupdict=self.mddataDict
            ),
        'x_charmm_frame_sequence_high_frequency_correction_kinetic_energy' : MetaInfoMap(startpage,
            activeSections=['section_frame_sequence'],
            depends=[{'store' : 'HFCKE'}],
            valtype='float',
            unitdict=self.unitDict,
            unit=self.akmaDict['energy'],
            lookupdict=self.mddataDict
            ),
        'x_charmm_frame_sequence_high_frequency_correction_energy_frames' : MetaInfoMap(startpage,
            activeSections=['section_frame_sequence'],
            depends=[{'store' : 'STEP'}],
            valtype='int',
            lookupdict=self.mddataDict
            ),
        'x_charmm_frame_sequence_high_frequency_correction_energy' : MetaInfoMap(startpage,
            activeSections=['section_frame_sequence'],
            depends=[{'store' : 'EHFCOR'}],
            valtype='float',
            unitdict=self.unitDict,
            unit=self.akmaDict['energy'],
            lookupdict=self.mddataDict
            ),
        'x_charmm_frame_sequence_virial_kinetic_energy_frames' : MetaInfoMap(startpage,
            activeSections=['section_frame_sequence'],
            depends=[{'store' : 'STEP'}],
            valtype='int',
            lookupdict=self.mddataDict
            ),
        'x_charmm_frame_sequence_virial_kinetic_energy' : MetaInfoMap(startpage,
            activeSections=['section_frame_sequence'],
            depends=[{'store' : 'VIRKE'}],
            valtype='float',
            unitdict=self.unitDict,
            unit=self.akmaDict['energy'],
            lookupdict=self.mddataDict
            ),
        #'x_charmm_frame_sequence_boundary_frames' : MetaInfoMap(startpage,
        #    activeSections=['section_frame_sequence'],
        #    depends=[{'store' : 'STEP'}],
        #    valtype='int',
        #    lookupdict=self.mddataDict
        #    ),
        #'x_charmm_frame_sequence_boundary' : MetaInfoMap(startpage,
        #    activeSections=['section_frame_sequence'],
        #    depends=[{'store' : 'BOUNDARY'}],
        #    valtype='float',
        #    unitdict=self.unitDict,
        #    unit='kilo-joule/mol',
        #    lookupdict=self.mddataDict
        #    ),
        'x_charmm_frame_sequence_bond_energy_frames' : MetaInfoMap(startpage,
            activeSections=['section_frame_sequence'],
            depends=[{'store' : 'STEP'}],
            valtype='int',
            lookupdict=self.mddataDict
            ),
        'x_charmm_frame_sequence_bond_energy' : MetaInfoMap(startpage,
            activeSections=['section_frame_sequence'],
            depends=[{'store' : 'BONDS'}],
            valtype='float',
            unitdict=self.unitDict,
            unit=self.akmaDict['energy'],
            lookupdict=self.mddataDict
            ),
        'x_charmm_frame_sequence_angle_energy_frames' : MetaInfoMap(startpage,
            activeSections=['section_frame_sequence'],
            depends=[{'store' : 'STEP'}],
            valtype='int',
            lookupdict=self.mddataDict
            ),
        'x_charmm_frame_sequence_angle_energy' : MetaInfoMap(startpage,
            activeSections=['section_frame_sequence'],
            depends=[{'store' : 'ANGLES'}],
            valtype='float',
            unitdict=self.unitDict,
            unit=self.akmaDict['energy'],
            lookupdict=self.mddataDict
            ),
        'x_charmm_frame_sequence_proper_dihedral_energy_frames' : MetaInfoMap(startpage,
            activeSections=['section_frame_sequence'],
            depends=[{'store' : 'STEP'}],
            valtype='int',
            lookupdict=self.mddataDict
            ),
        'x_charmm_frame_sequence_proper_dihedral_energy' : MetaInfoMap(startpage,
            activeSections=['section_frame_sequence'],
            depends=[{'store' : 'DIHEDRALS'}],
            valtype='float',
            unitdict=self.unitDict,
            unit=self.akmaDict['energy'],
            lookupdict=self.mddataDict
            ),
        'x_charmm_frame_sequence_improper_dihedral_energy_frames' : MetaInfoMap(startpage,
            activeSections=['section_frame_sequence'],
            depends=[{'store' : 'STEP'}],
            valtype='int',
            lookupdict=self.mddataDict
            ),
        'x_charmm_frame_sequence_improper_dihedral_energy' : MetaInfoMap(startpage,
            activeSections=['section_frame_sequence'],
            depends=[{'store' : 'IMPROPERS'}],
            valtype='float',
            unitdict=self.unitDict,
            unit=self.akmaDict['energy'],
            lookupdict=self.mddataDict
            ),
        'x_charmm_frame_sequence_cmap_dihedral_energy_frames' : MetaInfoMap(startpage,
            activeSections=['section_frame_sequence'],
            depends=[{'store' : 'STEP'}],
            valtype='int',
            lookupdict=self.mddataDict
            ),
        'x_charmm_frame_sequence_cmap_dihedral_energy' : MetaInfoMap(startpage,
            activeSections=['section_frame_sequence'],
            depends=[{'store' : 'CMAP'}],
            valtype='float',
            unitdict=self.unitDict,
            unit=self.akmaDict['energy'],
            lookupdict=self.mddataDict
            ),
        'x_charmm_frame_sequence_vdw_energy_frames' : MetaInfoMap(startpage,
            activeSections=['section_frame_sequence'],
            depends=[{'store' : 'STEP'}],
            valtype='int',
            lookupdict=self.mddataDict
            ),
        'x_charmm_frame_sequence_vdw_energy' : MetaInfoMap(startpage,
            activeSections=['section_frame_sequence'],
            depends=[{'store' : 'VDWAALS'}],
            valtype='float',
            unitdict=self.unitDict,
            unit=self.akmaDict['energy'],
            lookupdict=self.mddataDict
            ),
        'x_charmm_frame_sequence_electrostatic_energy_frames' : MetaInfoMap(startpage,
            activeSections=['section_frame_sequence'],
            depends=[{'store' : 'STEP'}],
            valtype='int',
            lookupdict=self.mddataDict
            ),
        'x_charmm_frame_sequence_electrostatic_energy' : MetaInfoMap(startpage,
            activeSections=['section_frame_sequence'],
            depends=[{'store' : 'ELEC'}],
            valtype='float',
            unitdict=self.unitDict,
            unit=self.akmaDict['energy'],
            lookupdict=self.mddataDict
            ),
        'x_charmm_frame_sequence_urey_energy_frames' : MetaInfoMap(startpage,
            activeSections=['section_frame_sequence'],
            depends=[{'store' : 'STEP'}],
            valtype='int',
            lookupdict=self.mddataDict
            ),
        'x_charmm_frame_sequence_urey_energy' : MetaInfoMap(startpage,
            activeSections=['section_frame_sequence'],
            depends=[{'store' : 'UREY-B'}],
            valtype='float',
            unitdict=self.unitDict,
            unit=self.akmaDict['energy'],
            lookupdict=self.mddataDict
            ),
        #'frame_sequence_to_sampling_method_ref' : MetaInfoMap(startpage),
        'geometry_optimization_converged' : MetaInfoMap(startpage,
            depends=[
                {'test' : [["MINI", " is not None"],
                           ['Minimization exiting with', ' is not None']],
                 'assign' : True},
                ],
            activeInfo=True,
            lookupdict=self.cntrlDict
            ),
        #'previous_sequence_ref' : MetaInfoMap(startpage)
        }

    frameseqend = {
        #'number_of_conserved_quantity_evaluations_in_sequence' : MetaInfoMap(startpage,
        #    activeInfo=True,
        #    value=(lambda x: np.array(x['val']).flatten().shape[0] if(
        #        x is not None and x['val'] is not None) else None)(
        #        self.metaStorage.fetchAttrValue(
        #            'xxx_to_rm_frame_sequence_conserved_quantity_frames'
        #            )
        #        )
        #    ),
        'number_of_frames_in_sequence' : MetaInfoMap(startpage, activeInfo=True,
            value=(lambda x: np.array(x['val']).flatten().shape[0] if(
                x is not None and x['val'] is not None) else None)(
                self.metaStorage.fetchAttrValue(
                    'xxx_to_rm_frame_sequence_potential_energy_frames'
                    )
                )
            ),
        'number_of_kinetic_energies_in_sequence' : MetaInfoMap(startpage, activeInfo=True,
            value=(lambda x: np.array(x['val']).flatten().shape[0] if(
                x is not None and x['val'] is not None) else None)(
                self.metaStorage.fetchAttrValue(
                    'xxx_to_rm_frame_sequence_kinetic_energy_frames'
                    )
                )
            ),
        'number_of_potential_energies_in_sequence' : MetaInfoMap(startpage,
            activeInfo=True,
            value=(lambda x: np.array(x['val']).flatten().shape[0] if(
                x is not None and x['val'] is not None) else None)(
                self.metaStorage.fetchAttrValue(
                    'xxx_to_rm_frame_sequence_potential_energy_frames'
                    )
                )
            ),
        'number_of_pressure_evaluations_in_sequence' : MetaInfoMap(startpage,
            activeInfo=True,
            value=(lambda x: np.array(x['val']).flatten().shape[0] if(
                x is not None and x['val'] is not None) else None)(
                self.metaStorage.fetchAttrValue(
                    'xxx_to_rm_frame_sequence_pressure_frames'
                    )
                )
            ),
        'number_of_temperatures_in_sequence' : MetaInfoMap(startpage,
            activeInfo=True,
            value=(lambda x: np.array(x['val']).flatten().shape[0] if(
                x is not None and x['val'] is not None) else None)(
                self.metaStorage.fetchAttrValue(
                    'xxx_to_rm_frame_sequence_temperature_frames'
                    )
                )
            ),
        'x_charmm_number_of_volumes_in_sequence' : MetaInfoMap(startpage,
            activeInfo=True,
            activeSections=['section_frame_sequence'],
            value=(lambda x: np.array(x['val']).flatten().shape[0] if(
                x is not None and x['val'] is not None) else None)(
                self.metaStorage.fetchAttrValue(
                    'x_charmm_frame_sequence_volume_frames'
                    )
                )
            ),
        'x_charmm_number_of_boundary_energies_in_sequence' : MetaInfoMap(startpage,
            activeInfo=True,
            activeSections=['section_frame_sequence'],
            value=(lambda x: np.array(x['val']).flatten().shape[0] if(
                x is not None and x['val'] is not None) else None)(
                self.metaStorage.fetchAttrValue(
                    'x_charmm_frame_sequence_boundary_frames'
                    )
                )
            ),
        'x_charmm_number_of_bond_energies_in_sequence' : MetaInfoMap(startpage,
            activeInfo=True,
            activeSections=['section_frame_sequence'],
            value=(lambda x: np.array(x['val']).flatten().shape[0] if(
                x is not None and x['val'] is not None) else None)(
                self.metaStorage.fetchAttrValue(
                    'x_charmm_frame_sequence_bond_energy_frames'
                    )
                )
            ),
        'x_charmm_number_of_angle_energies_in_sequence' : MetaInfoMap(startpage,
            activeInfo=True,
            activeSections=['section_frame_sequence'],
            value=(lambda x: np.array(x['val']).flatten().shape[0] if(
                x is not None and x['val'] is not None) else None)(
                self.metaStorage.fetchAttrValue(
                    'x_charmm_frame_sequence_angle_energy_frames'
                    )
                )
            ),
        'x_charmm_number_of_electrostatic_energies_in_sequence' : MetaInfoMap(startpage,
            activeInfo=True,
            activeSections=['section_frame_sequence'],
            value=(lambda x: np.array(x['val']).flatten().shape[0] if(
                x is not None and x['val'] is not None) else None)(
                self.metaStorage.fetchAttrValue(
                    'x_charmm_frame_sequence_electrostatic_energy_frames'
                    )
                )
            ),
        'x_charmm_number_of_vdw_energies_in_sequence' : MetaInfoMap(startpage,
            activeInfo=True,
            activeSections=['section_frame_sequence'],
            value=(lambda x: np.array(x['val']).flatten().shape[0] if(
                x is not None and x['val'] is not None) else None)(
                self.metaStorage.fetchAttrValue(
                    'x_charmm_frame_sequence_vdw_energy_frames'
                    )
                )
            ),
        'x_charmm_number_of_total2_energies_in_sequence' : MetaInfoMap(startpage,
            activeInfo=True,
            activeSections=['section_frame_sequence'],
            value=(lambda x: np.array(x['val']).flatten().shape[0] if(
                x is not None and x['val'] is not None) else None)(
                self.metaStorage.fetchAttrValue(
                    'x_charmm_frame_sequence_total2_energy_frames'
                    )
                )
            ),
        'x_charmm_number_of_total3_energies_in_sequence' : MetaInfoMap(startpage,
            activeInfo=True,
            activeSections=['section_frame_sequence'],
            value=(lambda x: np.array(x['val']).flatten().shape[0] if(
                x is not None and x['val'] is not None) else None)(
                self.metaStorage.fetchAttrValue(
                    'x_charmm_frame_sequence_total3_energy_frames'
                    )
                )
            ),
        'x_charmm_number_of_proper_dihedral_energies_in_sequence' : MetaInfoMap(startpage,
            activeInfo=True,
            activeSections=['section_frame_sequence'],
            value=(lambda x: np.array(x['val']).flatten().shape[0] if(
                x is not None and x['val'] is not None) else None)(
                self.metaStorage.fetchAttrValue(
                    'x_charmm_frame_sequence_proper_dihedral_energy_frames'
                    )
                )
            ),
        'x_charmm_number_of_improper_dihedral_energies_in_sequence' : MetaInfoMap(startpage,
            activeInfo=True,
            activeSections=['section_frame_sequence'],
            value=(lambda x: np.array(x['val']).flatten().shape[0] if(
                x is not None and x['val'] is not None) else None)(
                self.metaStorage.fetchAttrValue(
                    'x_charmm_frame_sequence_improper_dihedral_energy_frames'
                    )
                )
            ),
        'x_charmm_number_of_cmap_dihedral_energies_in_sequence' : MetaInfoMap(startpage,
            activeInfo=True,
            activeSections=['section_frame_sequence'],
            value=(lambda x: np.array(x['val']).flatten().shape[0] if(
                x is not None and x['val'] is not None) else None)(
                self.metaStorage.fetchAttrValue(
                    'x_charmm_frame_sequence_cmap_dihedral_energy_frames'
                    )
                )
            ),
        'x_charmm_number_of_misc_energies_in_sequence' : MetaInfoMap(startpage,
            activeInfo=True,
            activeSections=['section_frame_sequence'],
            value=(lambda x: np.array(x['val']).flatten().shape[0] if(
                x is not None and x['val'] is not None) else None)(
                self.metaStorage.fetchAttrValue(
                    'x_charmm_frame_sequence_misc_energy_frames'
                    )
                )
            ),
        #'previous_sequence_ref' : MetaInfoMap(startpage)
        }

    # ----------------------------------
    #   Definitions for section_system
    # ----------------------------------
    # section_system
    sec_system = {
        #'topology_ref' : MetaInfoMap(startpage),
        'atom_velocities' : MetaInfoMap(startpage),
        'configuration_raw_gid' : MetaInfoMap(startpage),
        'local_rotations' : MetaInfoMap(startpage),
        'number_of_atoms' : MetaInfoMap(startpage,
            depends=[{'value' : 'number_of_atoms'}],
            valtype='int',
            lookupdict=self.extraDict
            ),
        'number_of_sites' : MetaInfoMap(startpage),
        'number_of_symmetry_operations' : MetaInfoMap(startpage),
        'reduced_symmetry_matrices' : MetaInfoMap(startpage),
        'reduced_symmetry_translations' : MetaInfoMap(startpage),
        'sc_matrix' : MetaInfoMap(startpage),
        'spacegroup_3D_choice' : MetaInfoMap(startpage),
        'spacegroup_3D_hall' : MetaInfoMap(startpage),
        'spacegroup_3D_international' : MetaInfoMap(startpage),
        'spacegroup_3D_number' : MetaInfoMap(startpage),
        'spacegroup_3D_origin_shift' : MetaInfoMap(startpage),
        'spacegroup_3D_pointgroup' : MetaInfoMap(startpage),
        'spacegroup_3D_std_lattice' : MetaInfoMap(startpage),
        'spacegroup_3D_std_positions' : MetaInfoMap(startpage),
        'spacegroup_3D_std_types' : MetaInfoMap(startpage),
        'spacegroup_3D_trasformation_matrix' : MetaInfoMap(startpage),
        'spacegroup_3D_wyckoff' : MetaInfoMap(startpage),
        'symmorphic' : MetaInfoMap(startpage),
        'system_name' : MetaInfoMap(startpage,
            activeInfo=True,
            activeSections=['section_system'],
            subfunction={
                'function' : MDData.MDDataConverter.topology_system_name,
                'supportDict' : self.topoDict,
                },
            #functionbase=self
            ),
        'time_reversal_symmetry' : MetaInfoMap(startpage)
        }

    # section_configuration_core of section_system
    configuration_core = {
        'number_of_electrons' : MetaInfoMap(startpage,
            #value=0,
            #valtype='float',
            ),
        'atom_labels' : MetaInfoMap(startpage,
            #subfunction=self.system_atom_labels()
            ),
        'atom_positions' : MetaInfoMap(startpage,
            #subfunction=self.system_atom_positions()
            ),
        'configuration_periodic_dimensions' : MetaInfoMap(startpage,
            ),
        'embedded_system' : MetaInfoMap(startpage),
        'lattice_vectors' : MetaInfoMap(startpage,
            #subfunction=self.system_lattice_vectors()
            ),
        'simulation_cell' : MetaInfoMap(startpage,
            #subfunction=self.system_simulation_cell()
            )
        }

    # section_spacegroup_3D_operation of section_system
    spacegroup_op = {
            'spacegroup_3D_rotation' : MetaInfoMap(startpage),
            'spacegroup_3D_translation' : MetaInfoMap(startpage)
        }

    # section_system_to_system_refs of section_system
    sys_to_sys = {
            'system_to_system_kind' : MetaInfoMap(startpage),
            'system_to_system_ref' : MetaInfoMap(startpage)
        }

    # --------------------------------------------------------
    #   Definitions of meta data values for section_topology
    # --------------------------------------------------------
    # section_topology of section_run
    topology = {
        'atom_to_molecule' : MetaInfoMap(startpage,
            subfunction={
                'function' : MDData.MDDataConverter.topology_atom_to_mol,
                'supportDict' : self.topoDict,
                },
            #functionbase=self
            ),
        'molecule_to_molecule_type_map' : MetaInfoMap(startpage),
        'number_of_topology_atoms' : MetaInfoMap(startpage,
            depends=[{'value' : 'number_of_atoms'}],
            valtype='int',
            lookupdict=self.extraDict
            ),
        'number_of_topology_molecules' : MetaInfoMap(startpage,
            subfunction={
                'function' : MDData.MDDataConverter.topology_num_topo_mol,
                'supportDict' : self.topoDict,
                },
            valtype='int',
            ),
        'topology_force_field_name' : MetaInfoMap(startpage,
            depends=[{'value' : 'PARAMETER file'}],
            lookupdict=self.cntrlDict
            )
        }

    # section_atom_type of section_topology
    atom_type = {
        'atom_type_charge' : MetaInfoMap(startpage),
        'atom_type_mass' : MetaInfoMap(startpage),
        'atom_type_name' : MetaInfoMap(startpage)
        }

    # section_constraint of section_topology
    constraint = {
        'constraint_atoms' : MetaInfoMap(startpage),
        'constraint_kind' : MetaInfoMap(startpage),
        'constraint_parameters' : MetaInfoMap(startpage),
        'number_of_atoms_per_constraint' : MetaInfoMap(startpage),
        'number_of_constraints' : MetaInfoMap(startpage)
        }

    # section_interaction of section_topology
    interaction = {
        'interaction_atoms' : MetaInfoMap(startpage),
        'interaction_kind' : MetaInfoMap(startpage),
        'interaction_parameters' : MetaInfoMap(startpage),
        'number_of_atoms_per_interaction' : MetaInfoMap(startpage),
        'number_of_interactions' : MetaInfoMap(startpage)
        }

    # -------------------------------------------------------------
    #   Definitions of meta data values for section_molecule_type
    # -------------------------------------------------------------
    # section_molecule_type of section_topology
    mol_type = {
        'molecule_type_name' : MetaInfoMap(startpage),
        'number_of_atoms_in_molecule' : MetaInfoMap(startpage),
        'settings_atom_in_molecule' : MetaInfoMap(startpage)
        }

    # section_molecule_constraint of section_molecule_type
    mol_constraint = {
        'molecule_constraint_atoms' : MetaInfoMap(startpage),
        'molecule_constraint_kind' : MetaInfoMap(startpage),
        'molecule_constraint_parameters' : MetaInfoMap(startpage),
        'number_of_atoms_per_molecule_constraint' : MetaInfoMap(startpage),
        'number_of_molecule_constraints' : MetaInfoMap(startpage)
        }

    # section_molecule_interaction of section_molecule_type
    mol_interaction = {
        'molecule_interaction_atoms' : MetaInfoMap(startpage),
        'molecule_interaction_kind' : MetaInfoMap(startpage),
        'molecule_interaction_parameters' : MetaInfoMap(startpage),
        'number_of_atoms_per_molecule_interaction' : MetaInfoMap(startpage),
        'number_of_molecule_interactions' : MetaInfoMap(startpage)
        }

    # section_atom_in_molecule of section_molecule_type
    atom_in_mol = {
        'atom_in_molecule_charge' : MetaInfoMap(startpage),
        'atom_in_molecule_name' : MetaInfoMap(startpage),
        'atom_in_molecule_to_atom_type_ref' : MetaInfoMap(startpage)
        }


    if defname == 'system':
        dictionary = systemDict
    elif defname == 'topology':
        dictionary = topology
    elif defname == 'singleconfcalc':
        dictionary = singleconfcalc
    elif defname == 'restrictions':
        dictionary = restrictions
    elif defname == 'frameseq':
        dictionary = frameseq
    elif defname == 'frameseqend':
        dictionary = frameseqend
    elif defname == 'atom_type':
        dictionary = atom_type
    elif defname == 'molecule_type':
        dictionary = moltypeDict
    elif defname == 'interaction':
        dictionary = interaction
    elif defname == 'sampling':
        dictionary = sampling
    elif defname == 'singlevdw':
        dictionary = singlevdw
    elif defname == 'confcore':
        dictionary = configuration_core
    else:
        dictionary = singleconfcalclist
    return MapDictionary(dictionary)





# Copyright 2015-2018 Lauri Himanen, Fawzi Mohamed, Ankit Kariryaa
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import logging
import numpy as np
from nomadcore.baseclasses import AbstractBaseParser
logger = logging.getLogger("nomad")


class CP2KSinglePointForceParser(AbstractBaseParser):
    """Used to parse out a force file printed out by a CP2K single point
    calculation. It is not exactly an ZYX file, so here we define separate
    parser.
    """
    def __init__(self, parser_context):
        super(CP2KSinglePointForceParser, self).__init__(parser_context)

    def parse(self, filepath):
        start = False
        forces = []
        with open(filepath) as f:
            for line in f:
                if line.startswith(" # Atom"):
                    start = True
                    continue
                elif line.startswith(" SUM"):
                    forces = np.array(forces)
                    self.backend.addArrayValues("atom_forces", forces, unit="hartree / bohr")
                    break
                elif start:
                    components = [float(comp) for comp in line.split()[-3:]]
                    forces.append(components)



# Copyright 2018 Markus Scheidgen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an"AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

'''
This module provides function to establish connections to the database, searchengine, etc.
infrastructure services. Usually everything is setup at once with :func:`setup`. This
is run once for each *api* and *worker* process. Individual functions for partial setups
exist to facilitate testing, aspects of :py:mod:`nomad.cli`, etc.
'''

import os.path
import os
import shutil
from elasticsearch.exceptions import RequestError
from elasticsearch_dsl import connections
from mongoengine import connect, disconnect
from mongoengine.connection import ConnectionFailure
import smtplib
from email.mime.text import MIMEText
from keycloak import KeycloakOpenID, KeycloakAdmin
from keycloak.exceptions import KeycloakAuthenticationError, KeycloakGetError
import json
import jwt
from flask import g, request
import basicauth
from datetime import datetime
import re
import unidecode

from nomad import config, utils

logger = utils.get_logger(__name__)

elastic_client = None
''' The elastic search client. '''

mongo_client = None
''' The pymongo mongodb client. '''


def setup():
    '''
    Uses the current configuration (nomad/config.py and environment) to setup all the
    infrastructure services (repository db, mongo, elastic search) and logging.
    Will create client instances for the databases and has to be called before they
    can be used.
    '''
    setup_files()
    setup_mongo()
    setup_elastic()


def setup_files():
    for directory in [config.fs.public, config.fs.staging, config.fs.tmp]:
        if not os.path.exists(directory):
            os.makedirs(directory)


def setup_mongo(client=False):
    ''' Creates connection to mongodb. '''
    global mongo_client
    try:
        mongo_client = connect(db=config.mongo.db_name, host=config.mongo.host, port=config.mongo.port)
    except ConnectionFailure:
        disconnect()
        mongo_client = connect(db=config.mongo.db_name, host=config.mongo.host, port=config.mongo.port)

    logger.info('setup mongo connection')
    return mongo_client


def setup_elastic():
    ''' Creates connection to elastic search. '''
    global elastic_client
    elastic_client = connections.create_connection(
        hosts=['%s:%d' % (config.elastic.host, config.elastic.port)],
        timeout=60, max_retries=10, retry_on_timeout=True)
    logger.info('setup elastic connection')

    try:
        from nomad.search import entry_document
        entry_document.init(index=config.elastic.index_name)
    except RequestError as e:
        if e.status_code == 400 and 'resource_already_exists_exception' in e.error:
            # happens if two services try this at the same time
            pass
        else:
            raise e

    entry_document._index._name = config.elastic.index_name
    logger.info('initialized elastic index', index_name=config.elastic.index_name)

    return elastic_client


class Keycloak():
    '''
    A class that encapsulates all keycloak related functions for easier mocking and
    configuration
    '''
    def __init__(self):
        self.__oidc_client = None
        self.__admin_client = None
        self.__public_keys = None

    @property
    def _oidc_client(self):
        if self.__oidc_client is None:
            self.__oidc_client = KeycloakOpenID(
                server_url=config.keycloak.server_url,
                client_id=config.keycloak.client_id,
                realm_name=config.keycloak.realm_name,
                client_secret_key=config.keycloak.client_secret)

        return self.__oidc_client

    @property
    def _public_keys(self):
        if self.__public_keys is None:
            try:
                jwks = self._oidc_client.certs()
                self.__public_keys = {}
                for jwk in jwks['keys']:
                    kid = jwk['kid']
                    self.__public_keys[kid] = jwt.algorithms.RSAAlgorithm.from_jwk(
                        json.dumps(jwk))
            except Exception as e:
                self.__public_keys = None
                raise e

        return self.__public_keys

    def authorize_flask(self, basic: bool = True) -> str:
        '''
        Authorizes the current flask request with keycloak. Uses either Bearer or Basic
        authentication, depending on available headers in the request. Bearer auth is
        basically offline (besides retrieving and caching keycloaks public key for signature
        validation). Basic auth causes authentication agains keycloak with each request.

        Will set ``g.user``, either with None or user data from the respective OIDC token.

        Returns: An error message or None
        '''
        g.oidc_access_token = None
        if 'Authorization' in request.headers and request.headers['Authorization'].startswith('Bearer '):
            g.oidc_access_token = request.headers['Authorization'].split(None, 1)[1].strip()
        elif 'Authorization' in request.headers and request.headers['Authorization'].startswith('Basic '):
            if not basic:
                return 'Basic authentication not allowed, use Bearer token instead'

            try:
                auth = request.headers['Authorization'].split(None, 1)[1].strip()
                username, password = basicauth.decode(auth)
                token_info = self._oidc_client.token(username=username, password=password)
                g.oidc_access_token = token_info['access_token']
            except KeycloakAuthenticationError:
                return 'Could not authenticate, wrong credentials'
            except Exception as e:
                logger.error('Could not authenticate Basic auth', exc_info=e)
                return 'Could not authenticate Basic auth: %s' % str(e)

        if g.oidc_access_token is not None:
            auth_error: str = None
            try:
                kid = jwt.get_unverified_header(g.oidc_access_token)['kid']
                key = self._public_keys.get(kid)
                if key is None:
                    logger.error('The user provided keycloak public key does not exist. Does the UI use the right realm?')
                    auth_error = 'Could not verify JWT token: public key does not exist'
                else:
                    options = dict(verify_aud=False, verify_exp=True, verify_iss=True)
                    payload = jwt.decode(
                        g.oidc_access_token, key=key, algorithms=['RS256'], options=options,
                        issuer='%s/realms/%s' % (config.keycloak.server_url.rstrip('/'), config.keycloak.realm_name))

            except jwt.InvalidTokenError as e:
                auth_error = str(e)
            except Exception as e:
                logger.error('Could not verify JWT token', exc_info=e)
                raise e

            if auth_error is not None:
                g.user = None
                return auth_error

            else:
                from nomad import datamodel
                g.user = datamodel.User(
                    user_id=payload.get('sub', None),
                    email=payload.get('email', None),
                    first_name=payload.get('given_name', None),
                    last_name=payload.get('family_name', None))

                return None

        else:
            g.user = None
            # Do not return an error. This is the case were there are no credentials
            return None

    def __create_username(self, user):
        if user.first_name is not None and user.last_name is not None:
            user.username = '%s%s' % (user.first_name[:1], user.last_name)
        elif user.last_name is not None:
            user.username = user.last_name
        elif '@' in user.username:
            user.username = user.username.split('@')[0]

        user.username = unidecode.unidecode(user.username.lower())
        user.username = re.sub(r'[^0-9a-zA-Z_\-\.]+', '', user.username)

        index = 1
        try:
            while self.get_user(username=user.username):
                user.username += '%d' % index
                index += 1
        except KeyError:
            pass

    def add_user(self, user, bcrypt_password=None, invite=False):
        '''
        Adds the given :class:`nomad.datamodel.User` instance to the configured keycloak
        realm using the keycloak admin API.
        '''
        from nomad import datamodel
        if not isinstance(user, datamodel.User):
            if 'user_id' not in user:
                user['user_id'] = 'not set'

            if 'password' in user:
                bcrypt_password = user.pop('password')

            created = user.get('created', None)
            if created is not None and not isinstance(created, datetime):
                user['created'] = datetime.fromtimestamp(created / 1000)

            user = datamodel.User(**user)

        if user.username is None or not re.match(r'^[a-zA-Z0-9_\-\.]+$', user.username):
            self.__create_username(user)

        keycloak_user = dict(
            id=user.user_id if user.user_id != 'not set' else None,
            email=user.email,
            username=user.username,
            firstName=user.first_name,
            lastName=user.last_name,
            attributes=dict(
                repo_user_id=user.repo_user_id,
                affiliation=user.affiliation if user.affiliation is not None else '',
                affiliation_address=user.affiliation_address if user.affiliation_address is not None else ''),
            createdTimestamp=user.created.timestamp() * 1000 if user.created is not None else None,
            enabled=True,
            emailVerified=True)

        if invite:
            keycloak_user['requiredActions'] = [
                'UPDATE_PASSWORD', 'UPDATE_PROFILE', 'VERIFY_EMAIL']

        if bcrypt_password is not None:
            keycloak_user['credentials'] = [dict(
                type='password',
                hashedSaltedValue=bcrypt_password,
                algorithm='bcrypt')]

        keycloak_user = {
            key: value for key, value in keycloak_user.items()
            if value is not None}

        if user.user_id != 'not_set':
            try:
                self._admin_client.get_user(user.user_id)
                return 'User %s with given id already exists' % user.email
            except KeycloakGetError:
                pass

        if self._admin_client.get_user_id(user.email) is not None:
            return 'User with email %s already exists' % user.email

        try:
            self._admin_client.create_user(keycloak_user)
        except Exception as e:
            return str(e)

        if invite:
            try:
                user = self.get_user(username=user.username)
                self._admin_client.send_verify_email(user_id=user.user_id)
            except Exception as e:
                logger.error('could not send verify email', exc_info=e)

        return None

    def __user_from_keycloak_user(self, keycloak_user):
        from nomad import datamodel

        kwargs = {key: value[0] for key, value in keycloak_user.get('attributes', {}).items()}
        return datamodel.User(
            user_id=keycloak_user['id'],
            email=keycloak_user.get('email'),
            username=keycloak_user.get('username'),
            first_name=keycloak_user.get('firstName'),
            last_name=keycloak_user.get('lastName'),
            created=datetime.fromtimestamp(keycloak_user['createdTimestamp'] / 1000),
            **kwargs)

    def search_user(self, query: str = None, max=1000, **kwargs):
        if query is not None:
            kwargs['query'] = dict(search=query, max=max)
        else:
            kwargs['query'] = dict(max=max)
        try:
            keycloak_results = self._admin_client.get_users(**kwargs)
        except Exception as e:
            logger.error('Could not retrieve users from keycloak', exc_info=e)
            raise e

        return [
            self.__user_from_keycloak_user(keycloak_user)
            for keycloak_user in keycloak_results]

    def get_user(self, user_id: str = None, username: str = None, user=None) -> object:
        '''
        Retrives all available information about a user from the keycloak admin
        interface. This must be used to retrieve complete user information, because
        the info solely gathered from tokens (i.e. for the authenticated user ``g.user``)
        is generally incomplete.
        '''

        if user is not None and user_id is None:
            user_id = user.user_id

        if username is not None and user_id is None:
            with utils.lnr(logger, 'Could not use keycloak admin client'):
                user_id = self._admin_client.get_user_id(username)

            if user_id is None:
                raise KeyError('User with username %s does not exist' % username)

        assert user_id is not None, 'Could not determine user from given kwargs'

        try:
            keycloak_user = self._admin_client.get_user(user_id)

        except Exception as e:
            if str(getattr(e, 'response_code', 404)) == '404':
                raise KeyError('User does not exist')

            logger.error('Could not retrieve user from keycloak', exc_info=e)
            raise e

        return self.__user_from_keycloak_user(keycloak_user)

    @property
    def _admin_client(self):
        if True:  # TODO (self.__admin_client is None:), client becomes unusable after 60s
            self.__admin_client = KeycloakAdmin(
                server_url=config.keycloak.server_url,
                username=config.keycloak.username,
                password=config.keycloak.password,
                realm_name=config.keycloak.realm_name,
                verify=True)
            self.__admin_client.realm_name = config.keycloak.realm_name

        return self.__admin_client

    @property
    def access_token(self):
        return getattr(g, 'oidc_access_token', None)


keycloak = Keycloak()


def reset(remove: bool):
    '''
    Resets the databases mongo, elastic/calcs, and all files. Be careful.
    In contrast to :func:`remove`, it will only remove the contents of dbs and indicies.
    This function just attempts to remove everything, there is no exception handling
    or any warranty it will succeed.

    Args:
        remove: Do not try to recreate empty databases, remove entirely.
    '''
    try:
        if not mongo_client:
            setup_mongo()
        mongo_client.drop_database(config.mongo.db_name)
        logger.info('mongodb resetted')
    except Exception as e:
        logger.error('exception reset mongodb', exc_info=e)

    try:
        if not elastic_client:
            setup_elastic()
        elastic_client.indices.delete(index=config.elastic.index_name)
        from nomad.search import entry_document
        if not remove:
            entry_document.init(index=config.elastic.index_name)
        logger.info('elastic index resetted')
    except Exception as e:
        logger.error('exception resetting elastic', exc_info=e)

    try:
        shutil.rmtree(config.fs.staging, ignore_errors=True)
        shutil.rmtree(config.fs.public, ignore_errors=True)

        # delete tmp without the folder
        if os.path.isdir(config.fs.tmp):
            for sub_path in os.listdir(config.fs.tmp):
                path = os.path.join(config.fs.tmp, sub_path)
                try:
                    if os.path.isfile(path):
                        os.unlink(path)
                    elif os.path.isdir(path): shutil.rmtree(path, ignore_errors=True)
                except Exception:
                    pass

        logger.info('files resetted')
    except Exception as e:
        logger.error('exception deleting files', exc_info=e)


def send_mail(name: str, email: str, message: str, subject: str):
    """Used to programmatically send mails.

    Args:
        name: The email recipient name.
        email: The email recipient address.
        messsage: The email body.
        subject: The subject line.
    """
    if not config.mail.enabled:
        return

    logger = utils.get_logger(__name__)
    server = smtplib.SMTP(config.mail.host, config.mail.port)

    if config.mail.port == 995:
        try:
            server.starttls()
        except Exception as e:
            logger.warning('Could not use TTS', exc_info=e)

    if config.mail.with_login:
        try:
            server.login(config.mail.user, config.mail.password)
        except Exception as e:
            logger.warning('Could not log into mail server', exc_info=e)

    msg = MIMEText(message)
    msg['Subject'] = subject
    msg['To'] = name
    to_addrs = [email]

    if config.mail.cc_address is not None:
        msg['Cc'] = 'The nomad team <%s>' % config.mail.cc_address
        to_addrs.append(config.mail.cc_address)

    try:
        server.send_message(msg, from_addr=config.mail.from_address, to_addrs=to_addrs)
    except Exception as e:
        logger.error('Could not send email', exc_info=e)

    server.quit()



# Copyright 2016-2018 Ask Hjorth Larsen, Fawzi Mohamed
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

#
# Main author and maintainer: Ask Hjorth Larsen <asklarsen@gmail.com>

from __future__ import print_function
import os
import sys
import re

import numpy as np
import logging
from ase import Atoms
from ase.spacegroup import crystal
#from ase.data import chemical_symbols

from nomadcore.simple_parser import mainFunction, SimpleMatcher as SM
from nomadcore.local_meta_info import loadJsonFile, InfoKindEl
from nomadcore.unit_conversion.unit_conversion \
    import register_userdefined_quantity, convert_unit

# relative import
from gulpparser.spacegroups import get_spacegroup_number
#from util import floating, integer

# Generate list of all energy contributions:
# cat outputs/example*.got | grep "Components of energy" -A 20 | grep = |grep eV | cut -c 1-30 | sort|uniq|less
energy_contributions = """\
Attachment energy
Attachment energy/unit
Bond-order potentials
Brenner potentials
Bulk energy
Dispersion (real+recip)
Electric_field*distance
Energy shift
Four-body potentials
Improper torsions
Interatomic potentials
Many-body potentials
Monopole - monopole (real)
Monopole - monopole (recip)
Monopole - monopole (total)
Neutralising energy
  Non-primitive unit cell
Out of plane potentials
  Primitive unit cell
ReaxFF force field
Region 1-2 interaction
Region 2-2 interaction
Self energy (EEM/QEq/SM)
SM Coulomb correction
Solvation energy
Three-body potentials
Total lattice energy"""


def get_gulp_energy_patterns():
    for name in energy_contributions.splitlines():
        name = name.strip()
        basepattern = re.escape(name)
        name = name.lower()
        name = name.replace(' - ', '_')
        name = name.replace(' (', '_')
        name = name.replace(')', '')
        name = name.replace('field*distance', 'field_times_distance')
        name = re.sub(r'[-+/\s]', '_', name)

        metaname = 'x_gulp_energy_%s' % name
        pattern = r'\s*%s\s*=\s*(?P<%s__eV>\S+)\s*eV' % (basepattern,
                                                         metaname)
        yield name, metaname, pattern

energy_term_template = """{
    "description": "GULP energy term for %(name)s",
    "name": "%(metaname)s",
    "superNames": [ "section_single_configuration_calculation" ],
    "dtypeStr": "f",
    "shape": []
  }, """

def generate_energy_json():
    for name, metaname, pattern in get_gulp_energy_patterns():
        print(energy_term_template % dict(name=name,
                                          metaname=metaname), end='')
# generate_energy_json()  # Print JSON for energy terms to stdout

def get_gulp_energy_sm():
    sms = []
    for name, metaname, pattern in get_gulp_energy_patterns():
        #print(pattern)
        sms.append(SM(pattern, name=name))

    return SM(r'\s*Components of energy :',
              name='gulp-energy',
              subFlags=SM.SubFlags.Unordered,
              subMatchers=sms)

#TODO
#----

#OK    "program_name",
#OK    "atom_labels",
#OK    "atom_positions",
#OK    "program_version",
#OK    "energy_total",
#OK    "simulation_cell",
#OK    "configuration_periodic_dimensions"

#OK    'section_system',
#    'section_method',
#    'section_frame_sequence',
#    'section_sampling_method',
#OK    'single_configuration_to_calculation_method_ref',
#OK    'single_configuration_calculation_to_system_ref',
#    'atom_forces_raw',
#    'frame_sequence_local_frames_ref',
#    'frame_sequence_to_sampling_ref',

#DFT-only
#    'XC_functional_name',
#    'smearing_kind',
#    'smearing_width'
#    'eigenvalues_kpoints',
#    'eigenvalues_values',
#    'eigenvalues_occupation',
#    'band_k_points',
#    'band_energies',
#    'band_segm_start_end',
#    'band_segm_labels',
#    'dos_energies',
#    'dos_values',
#    'section_XC_functionals',
#    'program_basis_set_type',

"""File structure.  Example 4

Input for configuration = 1
  formula
  irreducible atoms/shells
  total atoms/shells
  dimensionalitycat
  symmetry
    space group
    cart lat vec
    cell param
    frac coord of asymm unit
  constraints

General input information
  species, type, atomic number, mass, charge, ... (table)
  general interatomic potentials (table)

"Start of fitting :"
  cycles... iterations
  fit completed successfuly
  final sum of squares
  final values of parameters (table)
  etc
  general interatomic potentials (table)
  total time to end of fitting

output for configuration 1
  components of energy :
    <components including total lattice energy>

"Start of bulk optimization"
  <iterations>
  Optimization achieved
  Final energy
  Components of energies <numbers>
  final asymmetric unit coords (table)
  final cartecian lattice vectors (ang) (table)
  final cell parameters and derivatives

comparison of initial and final structures: .....
time to end of optimisation
"job finished at ..."

"""


class GulpContext(object):
    def __init__(self):
        self.data = {}
        self.spacegroup = None
        self.current_raw_cell = None
        self.npbc = None
        self.use_spacegroup = None

        self.section_refs = {}

    def startedParsing(self, fname, parser):
        pass

    # Standard onOpen definitions from Siesta parser
    def onOpen_section_sampling_method(self, backend, gindex, section):
        self.section_refs['sampling_method'] = gindex

    def onOpen_section_frame_sequence(self, backend, gindex, section):
        self.section_refs['frame_sequence'] = gindex

    def onOpen_section_method(self, backend, gindex, section):
        self.section_refs['method'] = gindex

    def onOpen_section_system(self, backend, gindex, section):
        self.section_refs['system'] = gindex

    def adhoc_whether_using_spacegroup(self, parser):
        line = parser.fIn.readline()

        if 'Final coordinates of region' in line:
            self.use_spacegroup = False

    def onClose_section_method(self, backend, gindex, section):
        # species = self.data['species']
        # sym, types, charges = species.T
        # charges = convert_unit(charges.astype(float), 'e')
        # backend.addArrayValues('x_gulp_species_charge', charges)
        pass

    def onClose_section_system(self, backend, gindex, section):
        data = self.data

        ctable = data.pop('gulp_coordinate_table', None)
        if ctable is None:
            return  # This is probably an MD simulation or something.
            # I am sure we will get to suffer because of this relatively arbitrary
            # return, but what can you do

        symbols = ctable[:, 0]
        gulp_labels = ctable[:, 1]
        positions = ctable[:, 2:5].astype(float)

        chemical_symbols = []
        gulp_numeric_tags = []
        for sym in symbols:
            m = re.match(r'([A-Z][a-z]?)(\d*)', sym)
            chemical_symbols.append(m.group(1))
            gulp_numeric_tags.append(m.group(2))

        # Construct a mapping which combines gulp numeric tags and core/shell
        # etc. specification into a single numeric tag so ASE spacegroup
        # module can distinguish the atoms when constructing the full system
        types = {}
        tags = []

        ase_symbols = []
        for sym, label in zip(symbols, gulp_labels):
            full_label = '%s_%s' % (sym, label)
            tag = types.setdefault(full_label, len(types))
            tags.append(tag)
            ase_symbols.append(re.match(r'[A-Z][a-z]?', sym).group())

        # Really we want the opposite mapping
        tag2type = dict((v, k) for k, v in types.items())

        if self.npbc is None:
            self.npbc = section['x_gulp_pbc'][0]
        assert self.npbc in range(4)

        pbc = np.zeros(3, bool)
        pbc[:self.npbc] = True
        assert sum(pbc) == self.npbc

        # charge, occupancy?

        if self.spacegroup is None:
            spacegroup = section['x_gulp_space_group']
            if spacegroup is not None:
                spacegroup = spacegroup[0]
            self.spacegroup = spacegroup

        self.current_raw_cell = data.pop('cell', self.current_raw_cell)
        if self.npbc > 0:
            assert self.current_raw_cell is not None

        if self.use_spacegroup is None:
            self.use_spacegroup = (self.spacegroup is not None)

        if self.use_spacegroup:
            # group may be none ---- no spacegroup

            cellparnames = ['a', 'b', 'c', 'alpha', 'beta', 'gamma']

            #try:#if 'x_gulp_cell_a' in section:
            val = [section['x_gulp_cell_%s' % name] for name in cellparnames]
            if val[0] is None:
                # if we have the prim cell and not the full cell, it's because they
                # are identical (as far as I can tell).
                val = [section['x_gulp_prim_cell_%s' % name] for name in cellparnames]

            cellpar = []
            for v in val:
                assert len(v) == 1
                cellpar.append(v[0])
            assert len(cellpar) == 6

            num = get_spacegroup_number(self.spacegroup)

            basis_atoms = Atoms(ase_symbols, tags=tags,
                                cell=np.eye(3),
                                scaled_positions=positions)

            unique_labels = set(gulp_labels)
            thelabels = np.array(gulp_labels)

            constituents = []
            def build_atoms(label):
                b = basis_atoms[thelabels == label]
                atoms = crystal(b,
                                basis=b,
                                spacegroup=num,
                                cellpar=cellpar,
                                onduplicates='error')
                constituents.append(atoms)

            # Avoid duplicates when constructing space group atoms *grumble*
            # We want core and shell first
            for sym in 'cs':
                if sym in unique_labels:
                    unique_labels.remove(sym)
                    build_atoms(sym)
            for sym in sorted(unique_labels):
                build_atoms(sym)

            atoms = constituents[0]
            for other in constituents[1:]:
                atoms += other
        else:
            cell3d = np.identity(3)
            if self.npbc > 0:
                cell = self.current_raw_cell
                assert cell.shape == (self.npbc, 3), cell.shape
                cell3d[:self.npbc, :] = cell

            # use Atoms to get scaled positions/cell right:
            atoms = Atoms([0] * len(positions), cell=cell3d,
                          scaled_positions=positions, pbc=pbc,
                          tags=tags)

        atom_labels = [tag2type[tag] for tag in atoms.get_tags()]

        backend.addArrayValues('atom_labels', np.asarray(atom_labels))
        backend.addArrayValues('configuration_periodic_dimensions', pbc)
        backend.addValue('number_of_atoms', len(atom_labels))
        # The cell can be infinite in some directions.
        # In that case the cell value will just be one (it has to have some value!!)
        # but there will not be periodic boundary conditions in that direction.
        # We will have to live with this except if simulation_cell permits something
        # more general.
        backend.addArrayValues('simulation_cell',
                               convert_unit(atoms.cell, 'angstrom'))
        backend.addArrayValues('atom_positions',
                               convert_unit(atoms.positions, 'angstrom'))

    def onClose_section_single_configuration_calculation(self, backend,
                                                         gindex, section):
        backend.addValue('single_configuration_to_calculation_method_ref',
                         self.section_refs['method'])
        backend.addValue('single_configuration_calculation_to_system_ref',
                         self.section_refs['system'])

    # multi_sm is a copy from Siesta
    def multi_sm(self, name, startpattern, linepattern, endmatcher=None,
                 conflict='overwrite',  # 'fail', 'keep', 'overwrite'
                 *args, **kwargs):

        pat = re.compile(linepattern)  # XXX how to get compiled pattern?
        ngroups = pat.groups

        allgroups = []
        def addline(parser):
            line = parser.fIn.readline()
            match = pat.match(line)
            assert match is not None
            thislinegroups = match.groups()
            assert len(thislinegroups) == ngroups
            allgroups.append(thislinegroups)

        def savearray(parser):
            arr = np.array(allgroups, dtype=object)
            del allgroups[:]
            if name in self.data:
                if conflict == 'fail':
                    raise ValueError('grrr %s %s' % (name, self.data[name]))
                elif conflict == 'keep':
                    return  # Do not save array
                elif conflict == 'overwrite':
                    pass
                else:
                    raise ValueError('Unknown keyword %s' % conflict)
            #if arr.size > 0:
            if arr.size == 0:
                arr.shape = (0, ngroups)
            self.data[name] = arr

        if endmatcher is None:
            endmatcher = r'.*'

        if hasattr(endmatcher, 'swapcase'):
            endmatcher = SM(endmatcher,
                            endReStr='',
                            forwardMatch=True,
                            name='%s-end' % name,
                            adHoc=savearray)

        sm = SM(startpattern,
                name=name,
                subMatchers=[
                    SM(linepattern,
                       name='%s-line' % name,
                       repeats=True,
                       forwardMatch=True,
                       required=True,
                       adHoc=addline),
                    endmatcher,
                ], **kwargs)
        return sm


context = GulpContext()

def get_input_system_sm():
    m = SM(r'\*\s*Input for Configuration',
           name='input-conf',
           sections=['section_system'],
           subMatchers=[
               SM(r'\s*Formula\s*=\s*(?P<x_gulp_formula>\S+)',
                  name='formula'),
               SM(r'\s*Dimensionality\s*=\s*(?P<x_gulp_pbc>\d+)',
                  name='pbc'),
               SM(r'\s*Symmetry\s*:',
                  name='symm-header',
                  subMatchers=[
                      SM(r'\s*Space group \S+\s+:\s*(?P<x_gulp_space_group>.+?)\s*$',
                         name='spacegroup'),
                      SM(r'\s*Patterson group\s*:\s*(?P<x_gulp_patterson_group>.+?)\s*$',
                         name='patterson'),
                  ]),
               SM(r'\s*([Cc]artesian lattice|Surface [Cc]artesian|Polymer [Cc]artesian) vectors? \(Angstroms\) :',
                  name='lattice-header',
                  subMatchers=[
                      context.multi_sm('cell',
                                       r'',
                                       r'\s*(\S+)\s*(\S+)\s*(\S+)')
                  ]),
               SM(r'\s*Primitive cell parameters\s*:\s*Full cell parameters\s*:',
                  name='cellpar1',
                  subMatchers=[
                      SM(r'\s*a\s*=\s*\S+\s*alpha\s*=\s*\S+\s*a\s*=\s*(?P<x_gulp_cell_a>\S+)\s+alpha\s*=\s*(?P<x_gulp_cell_alpha>\S+)'),
                      SM(r'\s*b\s*=\s*\S+\s*beta\s*=\s*\S+\s*b =\s*(?P<x_gulp_cell_b>\S+)\s+beta\s*=\s*(?P<x_gulp_cell_beta>\S+)'),
                      SM(r'\s*c\s*=\s*\S+\s*gamma\s*=\s*\S+\s*c =\s*(?P<x_gulp_cell_c>\S+)\s+gamma\s*=\s*(?P<x_gulp_cell_gamma>\S+)'),
                  ]),
               SM(r'\s*Cell parameters\s*\(Angstroms/Degrees\):',
                  name='cellpar2',
                  subMatchers=[
                      SM(r'\s*a =\s*(?P<x_gulp_cell_a>\S+)\s*alpha\s*=\s*(?P<x_gulp_cell_alpha>\S+)'),
                      SM(r'\s*b =\s*(?P<x_gulp_cell_b>\S+)\s*beta\s*=\s*(?P<x_gulp_cell_beta>\S+)'),
                      SM(r'\s*c =\s*(?P<x_gulp_cell_c>\S+)\s*gamma\s*=\s*(?P<x_gulp_cell_gamma>\S+)'),
                  ]),
               SM(r'\s*(Fractional|[Cc]artesian|Mixed fractional/[Cc]artesian) coordinates of (asymmetric unit|surface|polymer|cluster|)\s*:',
                  #r'\s*(Fractional coordinates of asymmetric unit'
                  #r'|Mixed fractional/Cartesian coordinates of (surface|polymer)'
                  #r'|Cartesian coordinates of cluster)\s*:',
                  subFlags=SM.SubFlags.Sequenced,
                  name='frac-coords',
                  subMatchers=[
                      SM(r'--------+', name='bar'),
                      # We need to skip the 'Region 1' headers, so stop criterion is the empty line!
                      context.multi_sm('gulp_coordinate_table',
                                       r'-----+',
                                       #No.  Atomic       x           y          z         Charge      Occupancy
                                       #     Label      (Frac)      (Frac)     (Frac)        (e)         (Frac)
                                       #
                                       #     1      La      c       0.333333   0.666667   0.245000    *  9.00000    1.000000
                                       r'\s*\d+\s+(\S+)\s+(\S+)[\s\*]+(\S+)[\s\*]+(\S+)[\s\*]+(\S+)[\s\*]+(\S+)[\s\*]+(\S+)',
                                       r'------------+')
                  ]),
               #SM(r'General interatomic potentials',
               #),
            ])
    return m


def get_output_config_sm():
    m = SM(r'\*\s*Output for configuration',
           name='output-conf',
           endReStr='\s*Total lattice energy.*?kJ',
           sections=['section_single_configuration_calculation'],
           subMatchers=[
               get_gulp_energy_sm()
           ])
    return m

def get_optimise_sm():
    m = SM(r'\s*\*+\s*Optimisation achieved\s*\**',
        #r'\s*Start of \S+ optimisation :', #r'\*\s*Output for configuration',
           name='opt-start',
           repeats=True,
           sections=['section_system',
                     'section_single_configuration_calculation'],
           subMatchers=[
               get_gulp_energy_sm(),
               #SM(r'\s*Components of energy :',
               #   name='e-components',
               #   subMatchers=[
               #       get_gulp_energy_sm()
                      #SM(r'\s*Interatomic potentials\s*=\s*(?P<x_gulp_energy_interatomic_potentials__eV>\S+)\s*eV'),
                      #SM(r'\s*Monopole - monopole \(real\)\s*=\s*(?P<x_gulp_energy_monopole_monopole_real>\S+)\s*eV'),
                      #SM(r'\s*Monopole - monopole \(recip\)\s*=\s*(?P<x_gulp_energy_monopole_monopole_recip>\S+)\s*eV'),
                      #SM(r'\s*Monopole - monopole \(total\)\s*=\s*(?P<x_gulp_energy_monopole_monopole_total>\S+)\s*eV'),
                      #SM(r'\s*Electric_field\*distance\s*=\s*(?P<x_gulp_energy_electric_field>\S+)\s*eV'),
                      #SM(r'\s*Total lattice energy\s*=\s*(?P<energy_total__eV>\S+)\s*eV',
                      #   name='etotal')
               #   ]),
               SM(r'\s*Final (asymmetric unit|fractional|fractional/[Cc]artesian|[Cc]artesian)?\s*coordinates\s*(of atoms|of region \d+)?\s*:',
                  forwardMatch=True,
                  adHoc=context.adhoc_whether_using_spacegroup,
                  subMatchers=[
                      context.multi_sm('gulp_coordinate_table',
                                       r'-------------+',
                                       # This table is slightly differently formatted than the one from the input
                                       #--------------------------------------------------------------------------------
                                       #   No.  Atomic        x           y           z         Radius
                                       #        Label       (Frac)      (Frac)      (Frac)       (Angs)
                                       #--------------------------------------------------------------------------------
                                       #     1      La      c     0.333   0.666   0.248   0.000
                                       #     2      O       c     0.000   0.000   0.000   0.000
                                       r'\s*\d+\s*(\S+)\s*(\S+)\s*(\S+)\s*(\S+)\s*(\S+)\s*(\S+)',
                                       r'^$')
                  ]),
               SM(r'\s*Final [Cc]artesian lattice vectors \(Angstroms\) :',
                  subMatchers=[
                      context.multi_sm(r'cell',
                                       r'',
                                       r'\s*(\S+)\s*(\S+)\s*(\S+)')
                  ]),
               SM(r'\s*Final cell parameters and derivatives :',
                  name='finalcell',
                  subMatchers=[
                      SM(r'\s*a\s*(?P<x_gulp_prim_cell_a>\S+)\s+Angstrom', name='a'),
                      SM(r'\s*b\s*(?P<x_gulp_prim_cell_b>\S+)\s+Angstrom', name='b'),
                      SM(r'\s*c\s*(?P<x_gulp_prim_cell_c>\S+)\s+Angstrom', name='c'),
                      SM(r'\s*alpha\s*(?P<x_gulp_prim_cell_alpha>\S+)\s+Degrees', name='alpha'),
                      SM(r'\s*beta\s*(?P<x_gulp_prim_cell_beta>\S+)\s+Degrees', name='beta'),
                      SM(r'\s*gamma\s*(?P<x_gulp_prim_cell_gamma>\S+)\s+Degrees', name='gamma'),
                  ]),
               SM(r'\s*Non-primitive lattice parameters :',
                  name='nonprim',
                  subMatchers=[
                      SM(r'\s*a\s*=\s*(?P<x_gulp_cell_a>\S+)\s*b\s*=\s*(?P<x_gulp_cell_b>\S+)\s*c\s*=\s*(?P<x_gulp_cell_c>\S+)'),
                      SM(r'\s*alpha\s*=\s*(?P<x_gulp_cell_alpha>\S+)\s*beta\s*=\s*(?P<x_gulp_cell_beta>\S+)\s*gamma\s*=\s*(?P<x_gulp_cell_gamma>\S+)'),
                  ]),
           ])
    return m

def get_header_sm():
    m = SM(r'\*\s*GENERAL UTILITY LATTICE PROGRAM\s*\*',
           name='header',
           endReStr=r'\s*Job Started',
           subMatchers=[
               SM(r'\*\s+.*?\s*\*$',
                  name='Julian and friends',
                  repeats=True),
               SM(r'\*\*+'),
               SM(r'\*\s*Version\s*=\s*(?P<program_version>\S+)',
                  name='version'),
               SM(r'\*\*\*\*+',
                  endReStr=r'\*\*\*\*+',
                  name='bar',
                  subMatchers=[
                      SM(r'\*\s*(?P<x_gulp_main_keyword>\S+)\s*-',
                         repeats=True,
                         sections=['x_gulp_section_main_keyword'],
                         name='mainkw')
                  ]),
               SM(r'\*\s*(?P<x_gulp_title>.*?)\s*\*$',
                  name='title')
           ])
    return m

def get_gulp_potential_species_pattern(nspecies):
    tokens = []
    for i in range(1, nspecies + 1):
        tokens.append(r'(?P<x_gulp_forcefield_species_%d>\w+)\s*' % i)
        tokens.append(r'(?P<x_gulp_forcefield_speciestype_%d>\S+)\s*' % i)
    return ''.join(tokens)


def get_forcefield_table_sm(header, columnheaderpattern, tablepattern, name):
    m = SM(header,#r'\s*(General interatomic|Intramolecular|Intermolecular) potentials :',
           name=name,
           subMatchers=[
               SM(columnheaderpattern,#r'\s*Atom\s*Types\s*Potential\s*A*\s*B\s*C\s*D',
                  #Atom  Types   Potential         A         B         C         D     Cutoffs(Ang)
                  #  1     2                                                            Min    Max
                  # -------------------------------------------------------------------------------
                  #O    s La   s Buckingham    0.570E+04 0.299      38.9      0.00     0.000 24.000
                  subMatchers=[
                      SM(r'----------+', name='potentials',
                         endReStr=r'----------+',
                         subMatchers=[
                             SM(tablepattern,#''.join(tokens),
                                #get_gulp_potential_species_pattern(2) +
                                #r'(?P<x_gulp_forcefield_species_1>\w+)\s*'
                                #r'(?P<x_gulp_forcefield_speciestype_1>\S+)\s*'
                                #r'(?P<x_gulp_forcefield_species_2>\S+)\s*'
                                #r'(?P<x_gulp_forcefield_speciestype_2>\S+)\s*'
                                # The SRGlue potential is badly written in the table and unparseable.
                                # Probably a bug.  So just ignore it.
                                #r'(SRGlue|(?P<x_gulp_forcefield_potential_name>\b.{1,14}?)\s*'
                                #r'(?P<x_gulp_forcefield_parameter_a>\S+)\s*'
                                #r'(?P<x_gulp_forcefield_parameter_b>\S+)\s*'
                                #r'(?P<x_gulp_forcefield_parameter_c>\S+)\s*'
                               # r'(?P<x_gulp_forcefield_parameter_d>\S+)\s*'
                               # r'(?P<x_gulp_forcefield_cutoff_min>\S+\s*)'
                               # r'(?P<x_gulp_forcefield_cutoff_max>\S+))$',
                                name='table',
                                sections=['x_gulp_section_forcefield'],
                                repeats=True
                            ),
                         ]),
                  ]),
           ])
    return m


def get_general_input_sm():
    m = SM(r'\*\s*General input information',
           name='general-input',
           sections=['section_method'],
           subMatchers=[
               SM(r'\s*Species output for all configurations :'),
               SM(r'\s*Species\s*Type\s*Atomic',
                  subMatchers=[
                      context.multi_sm('species',
                                       r'-------+',
                                       #    species   type             charge
                                       r'\s*(\S+)\s*(\S+)\s*\d+\s*\S+\s*(\S+)',
                                       r'-------+')
                  ]),
               get_forcefield_table_sm(r'\s*(General interatomic|Intramolecular) potentials :',
                                       r'\s*Atom\s*Types\s*Potential\s*A*\s*B\s*C\s*D',
                                       get_gulp_potential_species_pattern(2) +
                                       r'(SRGlue|'
                                       r'(?P<x_gulp_forcefield_potential_name>\b.{1,14}?)\s*'
                                       r'(?P<x_gulp_forcefield_parameter_a>\S+)\s*'
                                       r'(?P<x_gulp_forcefield_parameter_b>\S+)\s*'
                                       r'(?P<x_gulp_forcefield_parameter_c>\S+)\s*'
                                       r'(?P<x_gulp_forcefield_parameter_d>\S+)\s*'
                                       r'(?P<x_gulp_forcefield_cutoff_min>\S+\s*)'
                                       r'(?P<x_gulp_forcefield_cutoff_max>\S+)$)',
                                       name='interatomic'),
               get_forcefield_table_sm(r'\s*(Intermolecular) potentials :',  # Basically a copy of the one just above
                                       r'\s*Atom\s*Types\s*Potential\s*A*\s*B\s*C\s*D',
                                       get_gulp_potential_species_pattern(2) +
                                       r'(SRGlue|(?P<x_gulp_forcefield_potential_name>\b.{1,14}?)\s*'
                                       r'(?P<x_gulp_forcefield_parameter_a>\S+)\s*'
                                       r'(?P<x_gulp_forcefield_parameter_b>\S+)\s*'
                                       r'(?P<x_gulp_forcefield_parameter_c>\S+)\s*'
                                       r'(?P<x_gulp_forcefield_parameter_d>\S+)\s*'
                                       r'(?P<x_gulp_forcefield_cutoff_min>\S+\s*)'
                                       r'(?P<x_gulp_forcefield_cutoff_max>\S+))$',
                                       name='intermolecular'),
               get_forcefield_table_sm(r'\s*General Three-body potentials :',
                                       r'\s*Atom\s*Atom\s*Atom\s*Force Constants\s*Theta',
                                       get_gulp_potential_species_pattern(3) +
                                       r'(?P<x_gulp_forcefield_threebody_1>\S+)\s*'
                                       r'(?P<x_gulp_forcefield_threebody_2>\S+)\s*'
                                       r'(?P<x_gulp_forcefield_threebody_3>\S+)\s*'
                                       r'(?P<x_gulp_forcefield_threebody_theta>\S+)\s*',
                                       name='3-body'),
               get_forcefield_table_sm(r'\s*General Four-body potentials :',
                                       r'\s*Atom Types\s*Force cst\s*Sign\s*Phase\s*Phi0',
                                       get_gulp_potential_species_pattern(4) +
                                       r'(?P<x_gulp_forcefield_fourbody_force_constant>\S+)\s*'
                                       r'(?P<x_gulp_forcefield_fourbody_sign>\S+)\s*'
                                       r'(?P<x_gulp_forcefield_fourbody_phase>\S+)\s*'
                                       r'(?P<x_gulp_forcefield_fourbody_phi0>\S+)\s*',
                                       name='3-body'),
           ])
    return m

def get_md_sm():
    m = SM(r'\*\s*Molecular Dynamics',
           name='md',
           subMatchers=[
               SM(r'\s*Molecular dynamics production :',
                  name='mdstep',
                  subMatchers=[
                      SM(r'\s*\*\*\s*Time :\s*(?P<x_gulp_md_time__ps>\S+)\s*ps :',
                         sections=['section_system', 'section_single_configuration_calculation'],
                         repeats=True,
                         subMatchers=[
                             SM(r'\s*Kinetic energy\s*\(eV\)\s*=\s*(?P<x_gulp_md_kinetic_energy__eV>\S+)'),
                             SM(r'\s*Potential energy\s*\(eV\)\s*=\s*(?P<x_gulp_md_potential_energy__eV>\S+)'),
                             SM(r'\s*Total energy\s*\(eV\)\s*=\s*(?P<x_gulp_md_total_energy__eV>\S+)'),
                             SM(r'\s*Temperature\s*\(K\)\s*=\s*(?P<x_gulp_md_temperature__K>\S+)'),
                             SM(r'\s*Pressure\s*\(GPa\)\s*=\s*(?P<x_gulp_md_pressure__GPa>\S+)'),
                         ])
                  ])
           ])
    return m

infoFileDescription = SM(
    name='root',
    weak=True,
    startReStr='',
    fixedStartValues={'program_name': 'gulp',
                      'program_basis_set_type': 'unavailable'},
    sections=['section_run'],
    subMatchers=[
        get_header_sm(),
        get_input_system_sm(),
        get_general_input_sm(),
        get_output_config_sm(),
        get_optimise_sm(),  # note British spelling
        get_md_sm(),
        SM(r'x^',
           name='impossible')  # 'Parse' the whole file
    ])


class GULPParser():
    """ A proper class envolop for running this parser from within python. """
    def __init__(self, backend, **kwargs):
        self.backend_factory = backend

    def parse(self, mainfile):
        from unittest.mock import patch
        logging.info('GULP parser started')
        logging.getLogger('nomadcore').setLevel(logging.WARNING)
        backend = self.backend_factory("gulp.nomadmetainfo.json")
        with patch.object(sys, 'argv', ['<exe>', '--uri', 'nmd://uri', mainfile]):
            mainFunction(
                infoFileDescription,
                None,
                {'name': 'gulp-parser', 'version': '1.0'},
                cachingLevelForMetaName={},
                superContext=GulpContext(),
                superBackend=backend)

        return backend



# Copyright 2016-2018 Ask Hjorth Larsen, Fawzi Mohamed
# 
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

#
# Main author and maintainer: Ask Hjorth Larsen <asklarsen@gmail.com>

varlist = [
    '%endblock',
    'Atom-Setup-Only',
    'Atom.Debug.KB.Generation',
    'AtomCoorFormatOut',
    'AtomLeftVcte',
    'AtomRightVcte',
    'AtomicCoordinatesFormat',
    'BasisPressure',
    'BornCharge',
    'BuildSuperCell',
    'BulkLead',
    'BulkTransport',
    'BulkTransvCellSize',
    'BulkTransvCellSizeX',
    'BulkTransvCellSizeY',
    'BulkTransvCellSizeZ',
    'CB.MaxKappa',
    'CB.WriteComplexBands',
    'CDFT',
    'CDFT.MaxIter',
    'COOP.Write',
    'CalcIETS',
    'CalcMPSH',
    'ChangeKgridInMD',
    'Compat-pre-v4-DM-H',
    'DM.AllowExtrapolation',
    'DM.AllowReuse',
    'DM.EnergyTolerance',
    'DM.FIRE.Mixing',
    'DM.FormattedFiles',
    'DM.FormattedInput',
    'DM.FormattedOutput',
    'DM.HarrisTolerance',
    'DM.KickMixingWeight',
    'DM.MixSCF1',
    'DM.MixingWeight',
    'DM.NormalizationTolerance',
    'DM.NormalizeDuringSCF',
    'DM.NumberBroyden',
    'DM.NumberKick',
    'DM.NumberPulay',
    'DM.OccupancyTolerance',
    'DM.Pulay.Avoid.First.After.Kick',
    'DM.PulayOnFile',
    'DM.RequireEnergyConvergence',
    'DM.RequireHarrisConvergence',
    'DM.Tolerance',
    'DM.UseSaveDM',
    'Delta',
    'DeltaWorkfunction',
    'Diag.AllInOne',
    'Diag.DivideAndConquer',
    'Diag.Memory',
    'Diag.NoExpert',
    'Diag.ParallelOverK',
    'Diag.PreRotate',
    'Diag.Use2D',
    'DiagMemory',
    'DiagScale',
    'DirectPhi',
    'DivideAndConquer',
    'EM.AddRhoGate',
    'EM.AddVgIsolatedLocalCharges',
    'EM.COOPCalculate',
    'EM.COOPNumberOfBonds',
    'EM.DebugRhoGate',
    'EM.NetRhoGateCharge',
    'EM.PrintLimits',
    'EM.RhoGateLxMax',
    'EM.RhoGateLxMin',
    'EM.RhoGateLyMax',
    'EM.RhoGateLyMin',
    'EM.RhoGateLzMax',
    'EM.RhoGateLzMin',
    'EM.TRCAddVCDFT',
    'EM.TimeReversal',
    'EM.Timings',
    'EM.addV',
    'EMPDOSKSO',
    'EMTransport',
    'ElectronicTemperature',
    'EnergLowestBound',
    'FilterCutoff',
    'FilterTol',
    'FinalTransmRange',
    'FixAuxiliaryCell',
    'FixAuxillaryCell',
    'FixSpin',
    'ForceAuxCell',
    'FullRamp',
    'HSetupOnly',
    'Harris_functional',
    'HartreeLeadsBottom',
    'HartreeLeadsLeft',
    'HartreeLeadsRight',
    'Ik_Select',
    'InitTransmRange',
    'InitTransport',
    'KB.New.Reference.Orbitals',
    'KB.Rmax',
    'LDAU.units',
    'LDAUForces',
    'LDAU_METHOD',
    'LatticeConstant',
    'LongOutput',
    'MD.AnnealOption',
    'MD.BulkModulus',
    'MD.FCAcousticSumRule',
    'MD.FCAtomRestart',
    'MD.FCAxisRestart',
    'MD.FCDispl',
    'MD.FCEigenVectors',
    'MD.FCIR',
    'MD.FCRead',
    'MD.FCfirst',
    'MD.FClast',
    'MD.FinalTimeStep',
    'MD.FireQuench',
    'MD.InitialTemperature',
    'MD.InitialTimeStep',
    'MD.LengthTimeStep',
    'MD.MaxCGDispl',
    'MD.MaxForceTol',
    'MD.MaxStressTol',
    'MD.NoseMass',
    'MD.NumCGsteps',
    'MD.NumNEBsteps',
    'MD.ParrinelloRahmanMass',
    'MD.Quench',
    'MD.RelaxCellOnly',
    'MD.RemoveIntraMolecularPressure',
    'MD.TRCSampling',
    'MD.TRCSkip',
    'MD.TargetPressure',
    'MD.TargetTemperature',
    'MD.TauRelax',
    'MD.Timing',
    'MD.TypeOfRun',
    'MD.UseSaveCG',
    'MD.UseSaveNEB',
    'MD.UseSaveXV',
    'MD.UseSaveZM',
    'MD.UseStructFile',
    'MD.VariableCell',
    'MM.Cutoff',
    'MM.UnitsDistance',
    'MM.UnitsEnergy',
    'MPSHAtomFirst',
    'MPSHAtomLast',
    'MPSHOrbFirst',
    'MPSHOrbLast',
    'MaxBondDistance',
    'MaxSCFIterations',
    'MeshCutoff',
    'MeshSubDivisions',
    'MinSCFIterations',
    'MixCharge',
    'MixHamiltonian',
    'MixedParallel',
    'MonitorForcesInSCF',
    'MullikenInSCF',
    'NC.OrbitalRotationEnd',
    'NC.OrbitalRotationStart',
    'NEB.SkipEdge',
    'NEnergReal',
    'NIVPoints',
    'NPoles',
    'NSlices',
    'NTransmPoints',
    'NaiveAuxiliaryCell',
    'NeglNonOverlapInt',
    'NenergImCircle',
    'NenergImLine',
    'NetCharge',
    'NonCollinearSpin',
    'NumSkipWriteDM',
    'NumberLinearMix',
    'NumberOfAtoms',
    'NumberOfEigenStates',
    'NumberOfSpecies',
    'Number_of_species',
    'ON.ChemicalPotential',
    'ON.ChemicalPotentialOrder',
    'ON.ChemicalPotentialRc',
    'ON.ChemicalPotentialTemperature',
    'ON.ChemicalPotentialUse',
    'ON.MaxNumIter',
    'ON.UseSaveLWF',
    'ON.eta',
    'ON.eta_alpha',
    'ON.eta_beta',
    'ON.etol',
    'ON.functional',
    'OccupationFunction',
    'On.RcLWF',
    'OpticalCalculation',
    'Optim.Broyden',
    'Output-Structure-Only',
    'PAO.BasisSize',
    'PAO.BasisType',
    'PAO.EnergyShift',
    'PAO.Filter',
    'PAO.FixSplitTable',
    'PAO.Keep.Findp.Bug',
    'PAO.NewSplitCode',
    'PAO.OldStylePolorbs',
    'PAO.SoftDefault',
    'PAO.SoftInnerRadius',
    'PAO.SoftPotential',
    'PAO.SplitNorm',
    'PAO.SplitNormH',
    'PAO.SplitTailNorm',
    'PS.SIC',
    'ParallelOverK',
    'PartialChargesAtEveryGeometry',
    'PartialChargesAtEveryScfStep',
    'PoissonMultigrid',
    'Print_ldau',
    'ProcessorGridX',
    'ProcessorGridY',
    'ProcessorGridZ',
    'ProjectionInSCF',
    'ProjectionMethod',
    'RcSpatial',
    'ReInitialiseDM',
    'ReadHamiltonian',
    'ReadKPIN',
    'ReparametrizePseudos',
    'Restricted.Radial.Grid',
    'Rmax.Radial.Grid',
    'RotateSpin.Phi',
    'RotateSpin.Theta',
    'SCF.LinearMixingAfterPulay',
    'SCF.MixAfterConvergence',
    'SCF.MixingWeightAfterPulay',
    'SCF.Pulay.Damping',
    'SCF.Pulay.DebugSVD',
    'SCF.Pulay.RcondSVD',
    'SCF.Pulay.UseSVD',
    'SCF.PulayDmaxRegion',
    'SCF.PulayMinimumHistory',
    'SCF.Read.Charge.NetCDF',
    'SCF.Read.Deformation.Charge.NetCDF',
    'SCF.Recompute-H-After-Scf',
    'SCF.Want.Variational.EKS',
    'SCFMustConverge',
    'SIC.Flavour',
    'SIC.Lambda',
    'SIC.NoRelaxation',
    'SIC.Npop',
    'SIC.PopDMConv',
    'SIC.PopKgridFactor',
    'SIC.PopSmatSparsity',
    'SIC.ProjectionMode',
    'SIC.ProjectionType',
    'SIC.Rot_Inv',
    'STT.Calculation',
    'STT.LinearResponse',
    'SaveBaderCharge',
    'SaveDeltaRho',
    'SaveElectrostaticPotential',
    'SaveHS',
    'SaveInitialChargeDensity',
    'SaveIonicCharge',
    'SaveNeutralAtomPotential',
    'SaveRho',
    'SaveRhoXC',
    'SaveTotalCharge',
    'SaveTotalPotential',
    'Scissor.Operator',
    'SetBulkTransvCell',
    'Siesta2Wannier90.NumberOfBands',
    'Siesta2Wannier90.NumberOfBandsDown',
    'Siesta2Wannier90.NumberOfBandsUp',
    'Siesta2Wannier90.WriteAmn',
    'Siesta2Wannier90.WriteEig',
    'Siesta2Wannier90.WriteMmn',
    'Siesta2Wannier90.WriteUnk',
    'Sigma.DSigmaDE',
    'SignatureRecords',
    'SimulateDoping',
    'SingleExcitation',
    'SkipLastIter',
    'SlabDipoleCorrection',
    'SolutionMethod',
    'SpinConfLeads',
    'SpinOrbit',
    'SpinPolarized',
    'SystemLabel',
    'SystemName',
    'TS.MixH',
    'TimeReversal',
    'TimeReversalSymmetryForKpoints',
    'TrCoefficients',
    'TryMemoryIncrease',
    'UseDomainDecomposition',
    'UseNewDiagk',
    'UseSaveData',
    'UseSpatialDecomposition',
    'UseStructFile',
    'UseTreeTimer',
    'VFinal',
    'VGate',
    'VInitial',
    'Vna.Filter',
    'WarningMinimumAtomicDistance',
    'WriteBands',
    'WriteCoorCerius',
    'WriteCoorInitial',
    'WriteCoorStep',
    'WriteCoorXmol',
    'WriteDM',
    'WriteDM.History.NetCDF',
    'WriteDM.NetCDF',
    'WriteDMHS.History.NetCDF',
    'WriteDMHS.NetCDF',
    'WriteDMT',
    'WriteDenchar',
    'WriteDiagdS',
    'WriteEDM',
    'WriteEigenvalues',
    'WriteForces',
    'WriteHSDeriv',
    'WriteHamiltonPop',
    'WriteHirshfeldPop',
    'WriteIonPlotFiles',
    'WriteKbands',
    'WriteKpoints',
    'WriteMDXmol',
    'WriteMDhistory',
    'WriteMullikenPop',
    'WriteProjections',
    'WriteSpinMulliken',
    'WriteSpinSCF',
    'WriteVNA',
    'WriteVoronoiPop',
    'WriteWaveFunctions',
    'XML.AbortOnErrors',
    'XML.AbortOnWarnings',
    'XML.Write',
    'ZBroadeningG',
    'ZLeftVcte',
    'ZM.CalcAllForces',
    'ZM.ForceTolAngle',
    'ZM.ForceTolLength',
    'ZM.MaxDisplAngle',
    'ZM.MaxDisplLength',
    'ZM.UnitsAngle',
    'ZM.UnitsLength',
    'ZRightVcte',
    'ZVGateL',
    'ZVGateR',
    'ZeemanTermBx',
    'ZeemanTermBy',
    'ZeemanTermBz',
    'alloc_report_level',
    'alloc_report_threshold',
    'blocksize',
    'compat-pre-v4-dynamics',
    'fdf-debug',
    'kgrid_cutoff',
    'processorY',
    'timer_report_threshold',
    'user-basis',
    'user-basis-netcdf',
    'xc.authors',
    'xc.functional',
]



# Copyright 2016-2018 Mikkel Strange, Fawzi Mohamed
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from __future__ import division
import os
from contextlib import contextmanager
import numpy as np
from ase.io.trajectory import Trajectory
from ase import units
from .constraint_conversion import get_nomad_name
from nomadcore.unit_conversion.unit_conversion import convert_unit as cu
import logging


@contextmanager
def open_section(p, name):
    gid = p.openSection(name)
    yield gid
    p.closeSection(name, gid)


def c(value, unit=None):
    """ Dummy function for unit conversion"""
#    return value
    return cu(value, unit)


parser_info = {"name": "parser_asap", "version": "1.0"}


def parse(filename, backend):
    t = Trajectory(filename, 'r')
    # some sanity checks
    if hasattr(t.backend, 'calculator'):
        if t.backend.calculator.get('name') != 'emt':  # Asap reports 'emt'!
            return

    if hasattr(t, 'description'):  # getting ready for improved traj format
        ds = t.description
    else:
        ds = {}

    p = backend
    o = open_section
    p.startedParsingSession(filename, parser_info)
    with o(p, 'section_run'):
        p.addValue('program_name', 'ASAP')
        if hasattr(t, 'ase_version') and t.ase_version:
            aversion = t.ase_version
        else:
            aversion = '3.x.x'  # default Asap version
        p.addValue('program_version', aversion)
        with o(p, 'section_topology'):
            ffn = t[0].calc.name  # maybe get it from asap3.todict method?
            p.addValue('topology_force_field_name', ffn)
            with o(p, 'section_constraint'):  # assuming constraints do not
                #indices = []                  # change from frame to frame
                for constraint in t[0].constraints:
                    d = constraint.todict()['kwargs']
                    if 'a' in d:
                        indices = np.array([d['a']])
                    else:
                        indices = d['indices']
                    p.addArrayValues('constraint_atoms',
                                     np.asarray(indices))
                    p.addValue('constraint_kind', get_nomad_name(constraint))
        with o(p, 'section_method') as method_gid:
            p.addValue('calculation_method', ffn)
        with o(p, 'section_frame_sequence'):
            for f in t:  # loop over frames
                with o(p, 'section_system') as system_gid:
                    p.addArrayValues('simulation_cell',
                                     c(f.get_cell(), 'angstrom'))
                    p.addArrayValues('atom_labels',
                                     np.asarray(f.get_chemical_symbols()))
                    p.addArrayValues('atom_positions',
                                     c(f.get_positions(), 'angstrom'))
                    p.addArrayValues('configuration_periodic_dimensions',
                                     f.get_pbc())
                    if f.get_velocities() is not None:
                        p.addArrayValues('atom_velocities',
                                         c(f.get_velocities() * units.fs /
                                           units.Angstrom,
                                           'angstrom/femtosecond'))
                with o(p, 'section_single_configuration_calculation'):
                    mref = 'single_configuration_to_calculation_method_ref'
                    sref = 'single_configuration_calculation_to_system_ref'
                    p.addValue(mref, method_gid)
                    p.addValue(sref, system_gid)
                    try:
                        p.addRealValue('energy_total',
                                    c(f.get_total_energy(), 'eV'))
                    except:
                        pass

                    try:
                        p.addArrayValues('atom_forces',
                                         c(f.get_forces(),
                                           'eV/angstrom'))
                    except:
                        pass

                    try:
                        p.addArrayValues('atom_forces_raw',
                                         c(f.get_forces(apply_constraint=False),
                                           'eV/angstrom'))
                    except:
                        pass
        with o(p, 'section_sampling_method'):
            ensemble_type = 'NVE'  # default ensemble_type
            if ds:  # if there is a traj.description
                if 'timestep' in ds:  # timestep in MD
                    p.addRealValue('x_asap_timestep', ds['timestep'])
                if 'maxstep' in ds:  # maxstep in relaxation
                    p.addRealValue('x_asap_maxstep', ds['maxstep'])
                if ds['type'] == 'optimization':
                    p.addValue('sampling_method', 'geometry_optimization')
                    p.addValue('geometry_optimization_method',
                               ds['optimizer'].lower())
                elif ds['type'] == 'molecular-dynamics':
                    p.addValue('sampling_method', 'molecular_dynamics')
                    try:
                        p.addRealValue('x_asap_temperature', ds['temperature'])
                    except:
                        pass
                    md_type = ds['md-type']
                    if 'Langevin' in md_type:
                        p.addValue('x_asap_langevin_friction', ds['friction'])
                        ensemble_type = 'NVT'
                    elif 'NVT' in md_type:
                        ensemble_type = 'NVT'
                    elif 'Verlet' in md_type:
                        ensemble_type = 'NVE'
                    elif 'NPT' in md_type:
                        ensemble_type = 'NPT'
            p.addValue('ensemble_type', ensemble_type)

    p.finishedParsingSession("ParseSuccess", None)


class AsapParser():
    """ A proper class envolop for running this parser from within python. """
    def __init__(self, backend, **kwargs):
        self.backend_factory = backend

    def parse(self, mainfile):
            logging.info('asap parser started')
            logging.getLogger('nomadcore').setLevel(logging.WARNING)
            backend = self.backend_factory("asap.nomadmetainfo.json")
            parserInfo = parser_info
            parse(mainfile, backend)
            return backend


# Copyright 2018 Markus Scheidgen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an"AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict
from elasticsearch_dsl import Q
from cachetools import cached

from optimade.filterparser import LarkParser
from optimade.filtertransformers.elasticsearch import ElasticTransformer, Quantity

from nomad.search import search_quantities


_parser = LarkParser(version=(0, 10, 1))


class FilterException(Exception):
    ''' Raised on parsing a filter expression with syntactic of semantic errors. '''
    pass


@cached(cache={})
def _get_transformer(nomad_properties, without_prefix):
    from nomad.datamodel import OptimadeEntry
    quantities: Dict[str, Quantity] = {
        q.name: Quantity(
            q.name, es_field='dft.optimade.%s' % q.name,
            elastic_mapping_type=q.a_search.mapping.__class__)

        for q in OptimadeEntry.m_def.all_quantities.values()
        if 'search' in q.m_annotations}

    quantities['elements'].length_quantity = quantities['nelements']
    quantities['dimension_types'].length_quantity = quantities['dimension_types']
    quantities['elements'].has_only_quantity = Quantity(name='only_atoms')
    quantities['elements'].nested_quantity = quantities['elements_ratios']
    quantities['elements_ratios'].nested_quantity = quantities['elements_ratios']

    if nomad_properties is not None:
        for search_quantity in search_quantities.values():
            name = search_quantity.name
            if '.' in name:
                if name.startswith(nomad_properties):
                    name = name[len(nomad_properties) + 1:]
                else:
                    continue

            names = ['_nmd_' + name]
            if without_prefix:
                names.append(name)

            for name in names:
                if name not in quantities:
                    quantities[name] = Quantity(
                        name,
                        es_field=search_quantity.search_field,
                        elastic_mapping_type=search_quantity.mapping.__class__)

    return ElasticTransformer(quantities=quantities.values())


def parse_filter(filter_str: str, nomad_properties='dft', without_prefix=False) -> Q:
    ''' Parses the given optimade filter str and returns a suitable elastic search query.

    Arguments:
        filter_str: Can be direct user input with no prior processing.
        nomad_properties: Also include the nomad proprietary properties of the given domain.
        without_prefix: Do not prefix the nomad proprietary properties with _nmd_.

    Raises:
        FilterException: If the given str cannot be parsed, or if there are any semantic
            errors in the given expression.
    '''
    transformer = _get_transformer(nomad_properties, without_prefix)

    try:
        parse_tree = _parser.parse(filter_str)
    except Exception as e:
        raise FilterException('Syntax error: %s' % str(e))

    try:
        query = transformer.transform(parse_tree)
    except Exception as e:
        raise FilterException('Semantic error: %s' % str(e))

    return query



import pytest

from optimade.adapters import Reference
from optimade.models import ReferenceResource


class TestReference:
    """Test Reference adapter"""

    def test_instantiate(self, RAW_REFERENCES):
        """Try instantiating Reference for all raw test references"""
        for reference in RAW_REFERENCES:
            new_Reference = Reference(reference)
            assert isinstance(new_Reference.entry, ReferenceResource)

    def test_setting_entry(self, capfd, RAW_REFERENCES):
        """Make sure entry can only be set once"""
        reference = Reference(RAW_REFERENCES[0])
        reference.entry = RAW_REFERENCES[1]
        captured = capfd.readouterr()
        assert "entry can only be set once and is already set." in captured.out

    @pytest.mark.skip(
        "Currently, there are no conversion types available for references"
    )
    def test_convert(self, reference):
        """Test convert() works
        Choose currently known entry type - must be updated if no longer available.
        """
        if not reference._type_converters:
            pytest.fail("_type_converters is seemingly empty. This should not be.")

        chosen_type = "SOME_VALID_TYPE"
        if chosen_type not in reference._type_converters:
            pytest.fail(
                f"{chosen_type} not found in _type_converters: {reference._type_converters} - "
                "please update test tests/adapters/references/test_references.py:TestReference."
                "test_convert()"
            )

        converted_reference = reference.convert(chosen_type)
        assert isinstance(converted_reference, (str, None.__class__))
        assert converted_reference == reference._converted[chosen_type]

    def test_convert_wrong_format(self, reference):
        """Test AttributeError is raised if format does not exist"""
        nonexistant_format = 0
        right_wrong_format_found = False
        while not right_wrong_format_found:
            if str(nonexistant_format) not in reference._type_converters:
                nonexistant_format = str(nonexistant_format)
                right_wrong_format_found = True
            else:
                nonexistant_format += 1

        with pytest.raises(
            AttributeError,
            match=f"Non-valid entry type to convert to: {nonexistant_format}",
        ):
            reference.convert(nonexistant_format)

    def test_getattr_order(self, reference):
        """The order of getting an attribute should be:
        1. `as_<entry type format>`
        2. `<entry type attribute>`
        3. `<entry type attributes attributes>`
        3. `raise AttributeError with custom message`
        """
        # If passing attribute starting with `as_`, it should call `self.convert()`
        with pytest.raises(
            AttributeError, match="Non-valid entry type to convert to: "
        ):
            reference.as_

        # If passing valid ReferenceResource attribute, it should return said attribute
        for attribute, attribute_type in (
            ("id", str),
            ("authors", list),
            ("attributes.authors", list),
        ):
            assert isinstance(getattr(reference, attribute), attribute_type)

        # Otherwise, it should raise AttributeError
        for attribute in ("nonexistant_attribute", "attributes.nonexistant_attribute"):
            with pytest.raises(AttributeError, match=f"Unknown attribute: {attribute}"):
                getattr(reference, attribute)
